
class InstrWCS(object):
    """
    A base class for instrument specific keyword definition.
    It prvides a default implementation (modeled by ACS) for
    all set_kw methods.
    """
    def __init__(self, hdr0=None, hdr=None):
        self.exthdr = hdr
        self.primhdr = hdr0
        self.set_ins_spec_kw()

    def set_ins_spec_kw(self):
        """
        This method MUST call all set_kw methods.
        There should be a set_kw method for all kw listed in
        mappings.ins_spec_kw. TypeError handles the case when
        fobj='DEFAULT'.
        """
        self.set_idctab()
        self.set_offtab()
        self.set_date_obs()
        self.set_ra_targ()
        self.set_dec_targ()
        self.set_pav3()
        self.set_detector()
        self.set_filter1()
        self.set_filter2()
        self.set_vafactor()
        self.set_naxis1()
        self.set_naxis2()
        self.set_ltv1()
        self.set_ltv2()
        self.set_binned()
        self.set_chip()
        self.set_parity()

    def set_idctab(self):
        try:
            self.idctab = self.primhdr['IDCTAB']
        except (KeyError, TypeError):
            self.idctab = None

    def set_offtab(self):
        try:
            self.offtab = self.primhdr['OFFTAB']
        except (KeyError, TypeError):
            self.offtab = None

    def set_date_obs(self):
        try:
            self.date_obs = self.primhdr['DATE-OBS']
        except (KeyError, TypeError):
            self.date_obs = None

    def set_ra_targ(self):
        try:
            self.ra_targ = self.primhdr['RA-TARG']
        except (KeyError, TypeError):
            self.ra_targ = None

    def set_dec_targ(self):
        try:
            self.dec_targ = self.primhdr['DEC-TARG']
        except (KeyError, TypeError):
            self.dec_targ = None

    def set_pav3(self):
        try:
            self.pav3 = self.primhdr['PA_V3']
        except (KeyError, TypeError):
            self.pav3 = None

    def set_filter1(self):
        try:
            self.filter1 = self.primhdr['FILTER1']
        except (KeyError, TypeError):
            self.filter1 = None

    def set_filter2(self):
        try:
            self.filter2 = self.primhdr['FILTER2']
        except (KeyError, TypeError):
            self.filter2 = None

    def set_vafactor(self):
        try:
            self.vafactor = self.exthdr['VAFACTOR']
        except (KeyError, TypeError):
            self.vafactor = 1

    def set_naxis1(self):
        try:
            self.naxis1 = self.exthdr['naxis1']
        except (KeyError, TypeError):
            try:
                self.naxis1 = self.exthdr['npix1']
            except (KeyError, TypeError):
                self.naxis1 = None

    def set_naxis2(self):
        try:
            self.naxis2 = self.exthdr['naxis2']
        except (KeyError, TypeError):
            try:
                self.naxis2 = self.exthdr['npix2']
            except (KeyError, TypeError):
                self.naxis2 = None

    def set_ltv1(self):
        try:
            self.ltv1 = self.exthdr['LTV1']
        except (KeyError, TypeError):
            self.ltv1 = 0.0

    def set_ltv2(self):
        try:
            self.ltv2 = self.exthdr['LTV2']
        except (KeyError, TypeError):
            self.ltv2 = 0.0

    def set_binned(self):
        try:
            self.binned = self.exthdr['BINAXIS1']
        except (KeyError, TypeError):
            self.binned = 1

    def set_chip(self):
        try:
            self.chip = self.exthdr['CCDCHIP']
        except (KeyError, TypeError):
            self.chip = 1

    def set_parity(self):
        self.parity = [[1.0, 0.0], [0.0, -1.0]]

    def set_detector(self):
        # each instrument has a different kw for detector and it can be
        # in a different header, so this is to be handled by the instrument classes
        self.detector = 'DEFAULT'


class ACSWCS(InstrWCS):
    """
    get instrument specific kw
    """

    def __init__(self, hdr0, hdr):
        self.primhdr = hdr0
        self.exthdr = hdr
        InstrWCS.__init__(self, hdr0, hdr)
        self.set_ins_spec_kw()

    def set_detector(self):
        try:
            self.detector = self.primhdr['DETECTOR']
        except KeyError:
            print('ERROR: Detector kw not found.\n')
            raise

    def set_parity(self):
        parity = {'WFC': [[1.0, 0.0], [0.0, -1.0]],
                  'HRC': [[-1.0, 0.0], [0.0, 1.0]],
                  'SBC': [[-1.0, 0.0], [0.0, 1.0]]}

        if self.detector not in list(parity.keys()):
            parity = InstrWCS.set_parity(self)
        else:
            self.parity = parity[self.detector]


class WFPC2WCS(InstrWCS):

    def __init__(self, hdr0, hdr):
        self.primhdr = hdr0
        self.exthdr = hdr
        InstrWCS.__init__(self, hdr0, hdr)
        self.set_ins_spec_kw()

    def set_filter1(self):
        self.filter1 = self.primhdr.get('FILTNAM1', None)
        if self.filter1 == " " or self.filter1 is None:
            self.filter1 = 'CLEAR1'

    def set_filter2(self):
        self.filter2 = self.primhdr.get('FILTNAM2', None)
        if self.filter2 == " " or self.filter2 is None:
            self.filter2 = 'CLEAR2'

    def set_binned(self):
        mode = self.primhdr.get('MODE', 1)
        if mode == 'FULL':
            self.binned = 1
        elif mode == 'AREA':
            self.binned = 2

    def set_chip(self):
        self.chip = self.exthdr.get('DETECTOR', 1)

    def set_parity(self):
        self.parity = [[-1.0, 0.], [0., 1.0]]

    def set_detector(self):
        try:
            self.detector = self.exthdr['DETECTOR']
        except KeyError:
            print('ERROR: Detector kw not found.\n')
            raise


class WFC3WCS(InstrWCS):
    """
    Create a WFC3 detector specific class
    """

    def __init__(self, hdr0, hdr):
        self.primhdr = hdr0
        self.exthdr = hdr
        InstrWCS.__init__(self, hdr0, hdr)
        self.set_ins_spec_kw()

    def set_detector(self):
        try:
            self.detector = self.primhdr['DETECTOR']
        except KeyError:
            print('ERROR: Detector kw not found.\n')
            raise

    def set_filter1(self):
        self.filter1 = self.primhdr.get('FILTER', None)
        if self.filter1 == " " or self.filter1 is None:
            self.filter1 = 'CLEAR'

    def set_filter2(self):
        # Nicmos idc tables do not allow 2 filters.
        self.filter2 = 'CLEAR'

    def set_parity(self):
        parity = {'UVIS': [[-1.0, 0.0], [0.0, 1.0]],
                  'IR': [[-1.0, 0.0], [0.0, 1.0]]}

        if self.detector not in list(parity.keys()):
            parity = InstrWCS.set_parity(self)
        else:
            self.parity = parity[self.detector]


class NICMOSWCS(InstrWCS):
    """
    Create a NICMOS specific class
    """

    def __init__(self, hdr0, hdr):
        self.primhdr = hdr0
        self.exthdr = hdr
        InstrWCS.__init__(self, hdr0, hdr)
        self.set_ins_spec_kw()

    def set_parity(self):
        self.parity = [[-1.0, 0.], [0., 1.0]]

    def set_filter1(self):
        self.filter1 = self.primhdr.get('FILTER', None)
        if self.filter1 == " " or self.filter1 is None:
            self.filter1 = 'CLEAR'

    def set_filter2(self):
        # Nicmos idc tables do not allow 2 filters.
        self.filter2 = 'CLEAR'

    def set_chip(self):
        self.chip = self.detector

    def set_detector(self):
        try:
            self.detector = self.primhdr['CAMERA']
        except KeyError:
            print('ERROR: Detector kw not found.\n')
            raise


class STISWCS(InstrWCS):
    """
    A STIS specific class
    """

    def __init__(self, hdr0, hdr):
        self.primhdr = hdr0
        self.exthdr = hdr
        InstrWCS.__init__(self, hdr0, hdr)
        self.set_ins_spec_kw()

    def set_parity(self):
        self.parity = [[-1.0, 0.], [0., 1.0]]

    def set_filter1(self):
        self.filter1 = self.exthdr.get('OPT_ELEM', None)
        if self.filter1 == " " or self.filter1 is None:
            self.filter1 = 'CLEAR1'

    def set_filter2(self):
        self.filter2 = self.exthdr.get('FILTER', None)
        if self.filter2 == " " or self.filter2 is None:
            self.filter2 = 'CLEAR2'

    def set_detector(self):
        try:
            self.detector = self.primhdr['DETECTOR']
        except KeyError:
            print('ERROR: Detector kw not found.\n')
            raise

    def set_date_obs(self):
        try:
            self.date_obs = self.exthdr['DATE-OBS']
        except (KeyError, TypeError):
            self.date_obs = None




import os
from stsci.tools import fileutil

import logging
logger = logging.getLogger("stwcs.updatewcs.utils")


def diff_angles(a, b):
    """
    Perform angle subtraction a-b taking into account
    small-angle differences across 360degree line.
    """
    return (a - b - 180.0) % 360.0 - 180.0


def getBinning(fobj, extver=1):
    # Return the binning factor
    if fobj[0].header['INSTRUME'] == 'WFPC2':
        mode = fobj[0].header.get('MODE', "")
        binned = 2 if mode == 'AREA' else 1
    else:
        binned = fobj['SCI', extver].header.get('BINAXIS', 1)
    return binned


def updateNEXTENDKw(fobj):
    """
    Updates PRIMARY header with correct value for NEXTEND, if present.

    Parameters
    -----------
    fobj : `astropy.io.fits.HDUList`
        The FITS object for file opened in `update` mode

    """
    if 'nextend' in fobj[0].header:
        fobj[0].header['nextend'] = len(fobj) - 1


def extract_rootname(kwvalue, suffix=""):
    """ Returns the rootname from a full reference filename

    If a non-valid value (any of ['','N/A','NONE','INDEF']) is input,
        simply return a string value of 'NONE'

    This function will also replace any 'suffix' specified with a blank.
    """
    # check to see whether a valid kwvalue has been provided as input
    if kwvalue.strip() in ['', 'N/A', 'NONE', 'INDEF']:
        return 'NONE'  # no valid value, so return 'NONE'

    # for a valid kwvalue, parse out the rootname
    # strip off any environment variable from input filename, if any are given
    if '$' in kwvalue:
        fullval = kwvalue[kwvalue.find('$') + 1:]
    else:
        fullval = kwvalue
    # Extract filename without path from kwvalue
    fname = os.path.basename(fullval).strip()

    # Now, rip out just the rootname from the full filename
    rootname = fileutil.buildNewRootname(fname)

    # Now, remove any known suffix from rootname
    rootname = rootname.replace(suffix, '')
    return rootname.strip()


def construct_distname(fobj, wcsobj):
    """
    This function constructs the value for the keyword 'DISTNAME'.
    It relies on the reference files specified by the keywords 'IDCTAB',
    'NPOLFILE', and 'D2IMFILE'.

    The final constructed value will be of the form:
        <idctab rootname>-<npolfile rootname>-<d2imfile rootname>
    and have a value of 'NONE' if no reference files are specified.
    """
    idcname = extract_rootname(fobj[0].header.get('IDCTAB', "NONE"),
                               suffix='_idc')
    if (idcname is None or idcname == 'NONE') and wcsobj.sip is not None:
        idcname = 'UNKNOWN'

    npolname, npolfile = build_npolname(fobj)
    if npolname is None and wcsobj.cpdis1 is not None:
        npolname = 'UNKNOWN'

    d2imname, d2imfile = build_d2imname(fobj)
    if d2imname is None and wcsobj.det2im is not None:
        d2imname = 'UNKNOWN'

    sipname, idctab = build_sipname(fobj)

    distname = build_distname(sipname, npolname, d2imname)
    return {'DISTNAME': distname, 'SIPNAME': sipname}


def build_distname(sipname, npolname, d2imname):
    """
    Core function to build DISTNAME keyword value without the HSTWCS input.
    """
    distname = sipname.strip()
    if npolname != 'NONE' or d2imname != 'NONE':
        if d2imname != 'NONE':
            distname += '-' + npolname.strip() + '-' + d2imname.strip()
        else:
            distname += '-' + npolname.strip()

    return distname


def build_default_wcsname(idctab):
    idcname = extract_rootname(idctab, suffix='_idc')
    wcsname = 'IDC_' + idcname
    return wcsname


def build_sipname(fobj, fname=None, sipname=None):
    """
    Build a SIPNAME from IDCTAB

    Parameters
    ----------
    fobj : `astropy.io.fits.HDUList`
        file object
    fname : string
          science file name (to be used if ROOTNAME is not present
    sipname : string
          user supplied SIPNAME keyword

    Returns
    -------
    sipname, idctab
    """
    try:
        idctab = fobj[0].header['IDCTAB']
        idcname = extract_rootname(idctab, suffix='_idc')
    except KeyError:
        idctab = 'N/A'
        idcname = 'N/A'
    if not fname:
        try:
            fname = fobj.filename()
        except Exception:
            fname = " "
    if not sipname:
        try:
            sipname = fobj[0].header["SIPNAME"]
            if idcname == 'N/A' or idcname not in sipname:
                raise KeyError
        except KeyError:
            if idcname != 'N/A':
                try:
                    rootname = fobj[0].header['rootname']
                except KeyError:
                    rootname = fname
                sipname = rootname + '_' + idcname
            else:
                if 'A_ORDER' in fobj[1].header or 'B_ORDER' in fobj[1].header:
                    sipname = 'UNKNOWN'
                else:
                    idcname = 'NOMODEL'

    return sipname, idctab


def build_npolname(fobj, npolfile=None):
    """
    Build a NPOLNAME from NPOLFILE

    Parameters
    ----------
    fobj : `astropy.io.fits.HDUList`
        file object
    npolfile : string
        user supplied NPOLFILE keyword

    Returns
    -------
    npolname, npolfile
    """
    if not npolfile:
        try:
            npolfile = fobj[0].header["NPOLFILE"]
        except KeyError:
            npolfile = ' '
            if fileutil.countExtn(fobj, 'WCSDVARR'):
                npolname = 'UNKNOWN'
            else:
                npolname = 'NOMODEL'
        npolname = extract_rootname(npolfile, suffix='_npl')
        if npolname == 'NONE':
            npolname = 'NOMODEL'
    else:
        npolname = extract_rootname(npolfile, suffix='_npl')
        if npolname == 'NONE':
            npolname = 'NOMODEL'
    return npolname, npolfile


def build_d2imname(fobj, d2imfile=None):
    """
    Build a D2IMNAME from D2IMFILE

    Parameters
    ----------
    fobj : `astropy.io.fits.HDUList`
        file object
    d2imfile : string
        user supplied NPOLFILE keyword

    Returns
    -------
    d2imname, d2imfile
    """
    if not d2imfile:
        try:
            d2imfile = fobj[0].header["D2IMFILE"]
        except KeyError:
            d2imfile = 'N/A'
            if fileutil.countExtn(fobj, 'D2IMARR'):
                d2imname = 'UNKNOWN'
            else:
                d2imname = 'NOMODEL'
        d2imname = extract_rootname(d2imfile, suffix='_d2i')
        if d2imname == 'NONE':
            d2imname = 'NOMODEL'
    else:
        d2imname = extract_rootname(d2imfile, suffix='_d2i')
        if d2imname == 'NONE':
            d2imname = 'NOMODEL'

    return d2imname, d2imfile


def remove_distortion(fname, dist_keyword):
    logger.info("Removing distortion {0} from file {0}"
                .format(dist_keyword, fname[0].header['rootname']))
    from ..wcsutil import altwcs
    if dist_keyword == "NPOLFILE":
        extname = "WCSDVARR"
        keywords = ["CPERR*", "DP1.*", "DP2.*", "CPDIS*", "NPOLEXT"]
    elif dist_keyword == "D2IMFILE":
        extname = "D2IMARR"
        keywords = ["D2IMERR*", "D2IM1.*", "D2IM2.*", "D2IMDIS*", "D2IMEXT"]
    else:
        raise AttributeError(
            "Unrecognized distortion keyword {} when attempting to remove "
            "distortion".format(dist_keyword)
        )
    ext_mapping = altwcs.mapFitsExt2HDUListInd(fname, "SCI").values()

    for hdu in ext_mapping:
        for kw in keywords:
            try:
                del fname[hdu].header[kw]
            except KeyError:
                pass
    ext_mapping = list(altwcs.mapFitsExt2HDUListInd(fname, extname).values())
    ext_mapping.sort()
    for hdu in ext_mapping[::-1]:
        del fname[hdu]



import shutil
import os
from astropy.io import fits as pyfits
from ..wcsutil import altwcs
from .. import updatewcs
from .. import wcsutil
import numpy as np
from numpy.testing import utils
import pytest

from . import data
data_path = os.path.split(os.path.abspath(data.__file__))[0]

os.environ['ASTROMETRY_STEP_CONTROL'] = 'Off'

def get_filepath(filename, directory=data_path):
    return os.path.join(directory, filename)


def compare_wcs(w1, w2, exclude_keywords=None):
    """
    Compare two WCSs.

    Parameters
    ----------
    w1, w2 : `astropy.wcs.WCS` objects
    exclude_keywords : list
        List of keywords to excude from comparison.
    """
    exclude_ctype = False
    keywords = ['crval', 'crpix', 'cd']
    if exclude_keywords is not None:
        exclude_keywords = [kw.lower() for kw in exclude_keywords]
        if 'ctype' in exclude_keywords:
            exclude_ctype = True
            exclude_keywords.remove('ctype')
        for kw in exclude_keywords:
            keywords.remove(kw)
    for kw in keywords:
        kw1 = getattr(w1.wcs, kw)
        kw2 = getattr(w2.wcs, kw)
        utils.assert_allclose(kw1, kw2, 1e-10)
    #utils.assert_allclose(w1.wcs.crpix, w2.wcs.crpix, 1e-10)
    #utils.assert_allclose(w1.wcs.cd, w2.wcs.cd, 1e-10)
    if not exclude_ctype:
        utils.assert_array_equal(np.array(w1.wcs.ctype), np.array(w2.wcs.ctype))

class TestAltWCS(object):

    def setup_class(self):
        acs_orig_file = get_filepath('j94f05bgq_flt.fits')
        simple_orig_file = get_filepath('simple.fits')
        current_dir = os.path.abspath(os.path.curdir)
        simple_file = get_filepath('simple.fits', current_dir)
        acs_file = get_filepath('j94f05bgq_flt.fits', current_dir)

        try:
            os.remove(acs_file)
            os.remove(simple_file)
        except OSError:
            pass
        idctab = get_filepath('postsm4_idc.fits')
        npol_file = get_filepath('qbu16424j_npl.fits')
        d2imfile = get_filepath('new_wfc_d2i.fits ')

        shutil.copyfile(acs_orig_file, acs_file)
        shutil.copyfile(simple_orig_file, simple_file)
        pyfits.setval(acs_file, ext=0, keyword="IDCTAB", value=idctab)
        pyfits.setval(acs_file, ext=0, keyword="NPOLFILE", value=npol_file)
        pyfits.setval(acs_file, ext=0, keyword="D2IMFILE", value=d2imfile)

        updatewcs.updatewcs(acs_file)
        self.acs_file = acs_file
        self.simplefits = simple_file
        self.ww = wcsutil.HSTWCS(self.acs_file, ext=1)

    def test_archive(self):
        altwcs.archiveWCS(self.acs_file, ext=1, wcskey='Z', wcsname='ZTEST', reusekey=False)
        w1 = wcsutil.HSTWCS(self.acs_file, ext=1)
        w1z = wcsutil.HSTWCS(self.acs_file, ext=1, wcskey='Z')
        compare_wcs(w1, w1z)

    def test_archive_clobber(self):
        altwcs.archiveWCS(self.acs_file, ext=1, wcskey='Z', wcsname='ZTEST', reusekey=True)
        w1 = wcsutil.HSTWCS(self.acs_file, ext=1)
        w1z = wcsutil.HSTWCS(self.acs_file, ext=1, wcskey='Z')
        compare_wcs(w1, w1z)

    def test_restore_wcs(self):
        # test restore on a file
        altwcs.restoreWCS(self.acs_file, ext=1, wcskey='O')
        w1o = wcsutil.HSTWCS(self.acs_file, ext=1, wcskey='O')
        w1 = wcsutil.HSTWCS(self.acs_file, ext=1)
        compare_wcs(w1, w1o, exclude_keywords=['ctype'])

    def test_restore_wcs_mem(self):
        # test restore on an HDUList object
        altwcs.archiveWCS(self.acs_file, ext=[('SCI', 1), ('SCI', 2)], wcskey='T')
        pyfits.setval(self.acs_file, ext=('SCI', 1), keyword='CRVAL1', value=1)
        pyfits.setval(self.acs_file, ext=('SCI', 2), keyword='CRVAL1', value=1)
        f = pyfits.open(self.acs_file, mode='update')
        altwcs.restoreWCS(f, ext=1, wcskey='T')
        f.close()
        w1o = wcsutil.HSTWCS(self.acs_file, ext=1, wcskey='T')
        w1 = wcsutil.HSTWCS(self.acs_file, ext=1)
        compare_wcs(w1, w1o)

    def test_restore_simple(self):
        # test restore on simple fits format
        altwcs.archiveWCS(self.simplefits, ext=0, wcskey='R')
        pyfits.setval(self.simplefits, ext=0, keyword='CRVAL1R', value=1)
        altwcs.restoreWCS(self.simplefits, ext=0, wcskey='R')
        wo = wcsutil.HSTWCS(self.simplefits, ext=0, wcskey='R')
        ws = wcsutil.HSTWCS(self.simplefits, ext=0)
        compare_wcs(ws, wo)

    def test_restore_wcs_from_to(self):
        # test restore from ... to ...
        pyfits.setval(self.acs_file, ext=('SCI', 1), keyword='CRVAL1', value=1)
        pyfits.setval(self.acs_file, ext=('SCI', 2), keyword='CRVAL1', value=1)
        f = pyfits.open(self.acs_file, mode='update')
        altwcs.restore_from_to(f, fromext='SCI', toext=['SCI', 'ERR', 'DQ'],
                               wcskey='T')
        f.close()
        w1o = wcsutil.HSTWCS(self.acs_file, ext=('SCI', 1), wcskey='T')
        w1 = wcsutil.HSTWCS(self.acs_file, ext=('SCI', 1))
        compare_wcs(w1, w1o)
        w2 = wcsutil.HSTWCS(self.acs_file, ext=('ERR', 1))
        compare_wcs(w2, w1o, exclude_keywords=['ctype'])
        w3 = wcsutil.HSTWCS(self.acs_file, ext=('DQ', 1))
        compare_wcs(w3, w1o, exclude_keywords=['ctype'])
        w4o = wcsutil.HSTWCS(self.acs_file, ext=4, wcskey='T')
        w4 = wcsutil.HSTWCS(self.acs_file, ext=('SCI', 2))
        compare_wcs(w4, w4o)
        w5 = wcsutil.HSTWCS(self.acs_file, ext=('ERR', 2))
        compare_wcs(w5, w4o, exclude_keywords=['ctype'])
        w6 = wcsutil.HSTWCS(self.acs_file, ext=('DQ', 2))
        compare_wcs(w3, w1o, exclude_keywords=['ctype'])

    def test_delete_wcs(self):
        altwcs.deleteWCS(self.acs_file, ext=1, wcskey='Z')
        with pytest.raises(KeyError):
            wcsutil.HSTWCS(self.acs_file, ext=1, wcskey='Z')

    def test_pars_file_mode1(self):
        assert(not altwcs._parpasscheck(self.acs_file, ext=1, wcskey='Z'))

    def test_pars_file_mode2(self):
        f = pyfits.open(self.acs_file)
        assert(not altwcs._parpasscheck(f, ext=1, wcskey='Z'))
        f.close()

    def test_pars_ext(self):
        f = pyfits.open(self.acs_file, mode='update')
        assert(altwcs._parpasscheck(f, ext=1, wcskey='Z'))
        assert(altwcs._parpasscheck(f, ext=[('sci', 1), ('sci', 2)], wcskey='Z'))
        assert(altwcs._parpasscheck(f, ext=('sci', 1), wcskey='Z'))
        f.close()

    def test_pars_wcskey_not1char(self):
        f = pyfits.open(self.acs_file, mode='update')
        assert(not altwcs._parpasscheck(f, ext=1, wcskey='ZZ'))
        f.close()

    def test_pars_wcskey(self):
        f = pyfits.open(self.acs_file, mode='update')
        assert(altwcs._parpasscheck(f, ext=1, wcskey=' '))
        #assert(not altwcs._parpasscheck(f, ext=1, wcskey=' ', reusekey=False))
        #assert(altwcs._parpasscheck(f, ext=1, wcskey='O'))
        #assert(not altwcs._parpasscheck(f, ext=1, wcskey='O', reusekey=False))
        f.close()



import shutil
import os
import io
from astropy.io import fits
from .. import updatewcs
from ..wcsutil import headerlet, wcsdiff
from ..wcsutil import HSTWCS
import numpy as np
from numpy.testing import utils
import pytest

from . import data
data_path = os.path.split(os.path.abspath(data.__file__))[0]


def get_filepath(filename, directory=data_path):
    return os.path.join(directory, filename)


class TestCreateHeaderlet(object):

    def setup_class(self):
        acs_orig_file = get_filepath('j94f05bgq_flt.fits')
        simple_orig_file = get_filepath('simple.fits')
        current_dir = os.path.abspath(os.path.curdir)
        simple_file = get_filepath('simple.fits', current_dir)
        acs_file = get_filepath('j94f05bgq_flt.fits', current_dir)
        comp_file = get_filepath('comp.fits', current_dir)
        self.headerlet_name = get_filepath('acs_hlet.fits', current_dir)

        bytes_file_orig = get_filepath('ia1d23dmq_test_hlet.txt')
        self.bytes_file = get_filepath('ia1d23dmq_test_hlet.txt', current_dir)
        self.bytes_hlet_file = get_filepath('ia1d23dmq_flt_hlet.fits')

        try:
            os.remove(acs_file)
            os.remove('comp.fits')
            os.remove(simple_file)
            os.remove(self.bytes_file)
        except OSError:
            pass
        idctab = get_filepath('postsm4_idc.fits')
        npol_file = get_filepath('qbu16424j_npl.fits')
        d2imfile = get_filepath('new_wfc_d2i.fits ')

        shutil.copyfile(acs_orig_file, acs_file)
        shutil.copyfile(simple_orig_file, simple_file)
        fits.setval(acs_file, ext=0, keyword="IDCTAB", value=idctab)
        fits.setval(acs_file, ext=0, keyword="NPOLFILE", value=npol_file)
        fits.setval(acs_file, ext=0, keyword="D2IMFILE", value=d2imfile)

        updatewcs.updatewcs(acs_file)

        shutil.copyfile(acs_file, comp_file)
        self.comp_file = comp_file
        self.simple_file = simple_file

        shutil.copyfile(bytes_file_orig, self.bytes_file)

    def testAllExt(self):
        """
        Test create_headerlet stepping through all
        extensions in the science file
        """
        hlet = headerlet.create_headerlet(self.comp_file, hdrname='hdr1')
        hlet.writeto(self.headerlet_name, overwrite=True)
        assert(wcsdiff.is_wcs_identical(self.comp_file, self.headerlet_name,
                                        [1, 4], [("SIPWCS", 1), ("SIPWCS", 2)],
                                        verbose=True)[0])

    def testSciExtList(self):
        """
        Test create_headerlet using a list of (EXTNAME, EXTNUM)
        extensions in the science file
        """
        hlet = headerlet.create_headerlet(self.comp_file,
                                          sciext=[('SCI', 1), ('SCI', 2)],
                                          hdrname='hdr1')
        hlet.writeto(self.headerlet_name, overwrite=True)
        assert(wcsdiff.is_wcs_identical(self.comp_file, self.headerlet_name,
                                        [1, 4], [("SIPWCS", 1), ("SIPWCS", 2)],
                                        verbose=True)[0])

    def testSciExtNumList(self):
        """
        Test create_headerlet using a list of EXTNUM
        extensions in the science file
        """
        hlet = headerlet.create_headerlet(self.comp_file,
                                          sciext=[1, 4],
                                          hdrname='hdr1')
        hlet.writeto(self.headerlet_name, overwrite=True)
        assert(wcsdiff.is_wcs_identical(self.comp_file, self.headerlet_name,
                                        [1, 4], [("SIPWCS", 1), ("SIPWCS", 2)],
                                        verbose=True)[0])

    def testHletFromSimpleFITS(self):
        """
        Test create_headerlet using a FITS HDUList extension
        number in the science file
        """
        hlet = headerlet.create_headerlet(self.simple_file,
                                          sciext=0,
                                          hdrname='hdr1')
        hlet.writeto(self.headerlet_name, overwrite=True)
        assert(wcsdiff.is_wcs_identical(self.simple_file, self.headerlet_name,
                                        [0], [1], verbose=True)[0])

    def test_no_HDRNAME_no_WCSNAME(self):
        """
        Test create_headerlet stepping through all
        extensions in the science file
        """
        newf = get_filepath('ncomp.fits', os.path.abspath(os.path.curdir))
        shutil.copyfile(self.comp_file, newf)
        #fits.delval(newf, 'HDRNAME', ext=1)
        fits.delval(newf, 'WCSNAME', ext=1)
        with pytest.raises(KeyError):
            hlet = headerlet.create_headerlet(newf)

    def test1SciExt(self):
        """
        Create a headerlet from only 1 SCI ext
        """
        hlet = headerlet.create_headerlet(self.comp_file,
                                          sciext=4,
                                          hdrname='hdr1')
        hlet.writeto(self.headerlet_name, overwrite=True)
        assert(wcsdiff.is_wcs_identical(self.comp_file, self.headerlet_name,
                                        [4], [1], verbose=True)[0])

    def testHletFromString(self):
        """Test creating a headerlet from a BytesIO object."""
        # read in test BytesIO object from test file
        hlet_comp = headerlet.Headerlet()
        hlet_comp = hlet_comp.fromfile(open(self.bytes_hlet_file, 'rb'))

        with open(self.bytes_file, 'rb') as f:
            hlet_bytes = io.BytesIO(f.read()).getvalue()

        hlet = headerlet.Headerlet(file=hlet_bytes)
        hlet = hlet.fromstring(hlet_bytes)  # fixed with PR#39

        assert(hlet_comp[0].header == hlet[0].header)
        assert(hlet_comp[0].data == hlet[0].data)
        assert(hlet_comp[1].header == hlet[1].header)
        assert(hlet_comp[1].data == hlet[1].data)

class TestApplyHeaderlet:

    def setup_class(self):
        acs_orig_file = get_filepath('j94f05bgq_flt.fits')
        simple_orig_file = get_filepath('simple.fits')
        current_dir = os.path.abspath(os.path.curdir)
        simple_file = get_filepath('simple.fits', current_dir)
        acs_file = get_filepath('j94f05bgq_flt.fits', current_dir)
        comp_file = get_filepath('comp.fits', current_dir)
        self.headerlet_name = get_filepath('acs_hlet.fits', current_dir)

        try:
            os.remove(acs_file)
            os.remove('comp.fits')
            os.remove(simple_file)
        except OSError:
            pass
        idctab = get_filepath('postsm4_idc.fits')
        npol_file = get_filepath('qbu16424j_npl.fits')
        d2imfile = get_filepath('new_wfc_d2i.fits ')

        shutil.copyfile(acs_orig_file, acs_file)
        shutil.copyfile(simple_orig_file, simple_file)
        fits.setval(acs_file, ext=0, keyword="IDCTAB", value=idctab)
        fits.setval(acs_file, ext=0, keyword="NPOLFILE", value=npol_file)
        fits.setval(acs_file, ext=0, keyword="D2IMFILE", value=d2imfile)

        updatewcs.updatewcs(acs_file)

        shutil.copyfile(acs_file, comp_file)
        self.comp_file = comp_file

    '''
    def setUp(self):
        try:
            os.remove('j94f05bgq_flt.fits')
            os.remove('comp.fits')
        except OSError:
            pass
        shutil.copyfile('orig/j94f05bgq_flt.fits', './j94f05bgq_flt.fits')
        updatewcs.updatewcs('j94f05bgq_flt.fits')
        shutil.copyfile('j94f05bgq_flt.fits', './comp.fits')
    '''
    """
    Does not raise an error currently, should it?
    @raises(TypeError)
    def testWrongDestim(self):
        hlet = headerlet.create_headerlet('comp.fits', sciext=4,
                                          hdrname='hdr1', destim='WRONG')
        hlet.apply_as_primary('comp.fits')
    """


    def testWrongSIPModel(self):
        hlet = headerlet.create_headerlet(self.comp_file, hdrname='test1',
                                          sipname='WRONG')
        with pytest.raises(ValueError):
            hlet.apply_as_primary(self.comp_file)

    def testWrongNPOLModel(self):
        hlet = headerlet.create_headerlet(self.comp_file, hdrname='test1',
                                          npolfile='WRONG')
        with pytest.raises(ValueError):
            hlet.apply_as_primary(self.comp_file)

    def testWrongD2IMModel(self):
        hlet = headerlet.create_headerlet(self.comp_file, hdrname='test1',
                                          d2imfile='WRONG')
        with pytest.raises(ValueError):
            hlet.apply_as_primary(self.comp_file)

    def test_apply_as_primary_method(self):
        hlet = headerlet.create_headerlet(self.comp_file, hdrname='test2')
        hlet['SIPWCS', 1].header['CRPIX1'] = 1
        hlet['SIPWCS', 1].header['CRPIX2'] = 1
        hlet['SIPWCS', 2].header['CRPIX1'] = 2
        hlet['SIPWCS', 2].header['CRPIX2'] = 2
        hlet.apply_as_primary(self.comp_file)
        hlet.writeto(self.headerlet_name, overwrite=True)
        assert(wcsdiff.is_wcs_identical(self.comp_file, self.headerlet_name,
                                        [('SCI', 1), ('SCI', 2)],
                                        [("SIPWCS", 1), ("SIPWCS", 2)],
                                        verbose=True)[0])
        # repeated hlet should not change sci file
        try:
            headerlet.apply_headerlet_as_primary(self.comp_file, 'hdr1_hlet.fits')
        except:
            pass
        assert(wcsdiff.is_wcs_identical(self.comp_file, self.headerlet_name,
                                        [('SCI', 1), ('SCI', 2)],
                                        [("SIPWCS", 1), ("SIPWCS", 2)],
                                        verbose=True)[0])

    def test_apply_as_alternate_method(self):
        hlet = headerlet.create_headerlet(self.comp_file, hdrname='test1')
        hlet.apply_as_alternate(self.comp_file, wcskey='K', wcsname='KK')
        hlet.writeto(self.headerlet_name, overwrite=True)
        assert(wcsdiff.is_wcs_identical(self.comp_file, self.headerlet_name,
                                        [('SCI', 1), ('SCI', 2)],
                                        [("SIPWCS", 1), ("SIPWCS", 2)],
                                        scikey='K', verbose=True)[0])
        headerlet.apply_headerlet_as_alternate(self.comp_file,
                                               self.headerlet_name, wcskey='P')
        assert(wcsdiff.is_wcs_identical(self.comp_file, self.headerlet_name,
                                        [('SCI', 1), ('SCI', 2)],
                                        [("SIPWCS", 1), ("SIPWCS", 2)],
                                        scikey='P', verbose=True)[0])

class TestRestoreHeaderlet:

    def setup_class(self):
        acs_orig_file = get_filepath('j94f05bgq_flt.fits')
        current_dir = os.path.abspath(os.path.curdir)
        acs_file = get_filepath('j94f05bgq_flt.fits', current_dir)
        self.headerlet_name = get_filepath('acs_hlet.fits', current_dir)

        try:
            os.remove(acs_file)
            os.remove('comp.fits')
        except OSError:
            pass

        idctab = get_filepath('postsm4_idc.fits')
        npol_file = get_filepath('qbu16424j_npl.fits')
        d2imfile = get_filepath('new_wfc_d2i.fits ')

        shutil.copyfile(acs_orig_file, acs_file)
        fits.setval(acs_file, ext=0, keyword="IDCTAB", value=idctab)
        fits.setval(acs_file, ext=0, keyword="NPOLFILE", value=npol_file)
        fits.setval(acs_file, ext=0, keyword="D2IMFILE", value=d2imfile)

        updatewcs.updatewcs(acs_file)
        self.sci_file = acs_file

    def test_restore_headerlet(self):
        hdrname = 'test1'
        hlet = headerlet.create_headerlet(self.sci_file, hdrname='test1')
        # modify hlet so that it is a unique solution
        hlet['sipwcs',1].header['crval1'] += 1./3600.
        hlet['sipwcs',1].header['crval2'] += 1./3600.
        hlet['sipwcs',2].header['crval1'] += 1./3600.
        hlet['sipwcs',2].header['crval2'] += 1./3600.
        hlet.writeto(self.headerlet_name, overwrite=True)

        headerlet.attach_headerlet(self.sci_file, self.headerlet_name)
        hlet_extn = headerlet.find_headerlet_HDUs(self.sci_file,
                                                  hdrname=hdrname)[0]
        headerlet.restore_from_headerlet(self.sci_file, hdrext=hlet_extn)

        assert(wcsdiff.is_wcs_identical(self.sci_file, self.headerlet_name,
                                        [('SCI', 1), ('SCI', 2)],
                                        [("SIPWCS", 1), ("SIPWCS", 2)],
                                        scikey=' ', verbose=True)[0])


class TestLegacyFiles:

    def setup_class(self):
        acs_orig_file = get_filepath('j94f05bgq_flt.fits')
        current_dir = os.path.abspath(os.path.curdir)
        self.acs_file = get_filepath('j94f05bgq_flt.fits', current_dir)
        self.legacy_file = get_filepath('jlegacy.fits', current_dir)

        try:
            os.remove(self.acs_file)
            os.remove(self.legacy_file)
        except OSError:
            pass
        idctab = get_filepath('postsm4_idc.fits')
        npol_file = get_filepath('qbu16424j_npl.fits')
        d2imfile = get_filepath('new_wfc_d2i.fits ')

        shutil.copyfile(acs_orig_file, self.acs_file)
        shutil.copyfile(acs_orig_file, self.legacy_file)

        fits.setval(self.acs_file, ext=0, keyword="IDCTAB", value=idctab)
        fits.setval(self.acs_file, ext=0, keyword="NPOLFILE", value=npol_file)
        fits.setval(self.acs_file, ext=0, keyword="D2IMFILE", value=d2imfile)
        updatewcs.updatewcs(self.acs_file)

    '''
    def setUp(self):
        try:
            os.remove('j94f05bgq_flt.fits')
            os.remove('jlegacy.fits')
        except OSError:
            pass
        shutil.copyfile('orig/j94f05bgq_flt.fits', './j94f05bgq_flt.fits')
        shutil.copyfile('j94f05bgq_flt.fits', './jlegacy.fits')
        updatewcs.updatewcs('j94f05bgq_flt.fits')
    '''

    def test_update_legacy_file(self):
        hlet = headerlet.create_headerlet(self.acs_file)
        hlet.apply_as_primary(self.legacy_file, archive=False, attach=False)
        assert(wcsdiff.is_wcs_identical(self.acs_file, self.legacy_file,
                                        [('SCI', 1), ('SCI', 2)],
                                        [("SCI", 1), ("SCI", 2)],
                                        verbose=True)[0])



import numpy as np
from matplotlib import pyplot as plt
from astropy.io import fits
import string

from stsci.tools import parseinput, irafglob
from ..distortion import utils
from .. import wcsutil
from ..wcsutil import altwcs


def vmosaic(fnames, outwcs=None, ref_wcs=None, ext=None, extname=None, undistort=True,
            wkey='V', wname='VirtualMosaic', plot=False, clobber=False):
    """
    Create a virtual mosaic using the WCS of the input images.

    Parameters
    ----------
    fnames: a string or a list
              a string or a list of filenames, or a list of wcsutil.HSTWCS objects
    outwcs: an HSTWCS object
             if given, represents the output tangent plane
             if None, the output WCS is calculated from the input observations.
    ref_wcs: an HSTWCS object
             if output wcs is not given, this will be used as a reference for the
             calculation of the output WCS. If ref_wcs is None and outwcs is None,
             then the first observation in th einput list is used as reference.
    ext:    an int, a tuple or a list
              an int - represents a FITS extension, e.g. 0 is the primary HDU
              a tuple - uses the notation (extname, extver), e.g. ('sci',1)
              Can be a list of integers or tuples representing FITS extensions
    extname: string
              the value of the EXTNAME keyword for the extensions to be used in the mosaic
    undistort: boolean (default: True)
               undistort (or not) the output WCS
    wkey:   string
              default: 'V'
              one character A-Z to be used to record the virtual mosaic WCS as
              an alternate WCS in the headers of the input files.
    wname:  string
              default: 'VirtualMosaic
              a string to be used as a WCSNAME value for the alternate WCS representign
              the virtual mosaic
    plot:   boolean
              if True and matplotlib is installed will make a plot of the tangent plane
              and the location of the input observations.
    clobber: boolean
              This covers the case when an alternate WCS with the requested key
              already exists in the header of the input files.
              if clobber is True, it will be overwritten
              if False, it will compute the new one but will not write it to the headers.

    Notes
    -----
    The algorithm is:
    1. If output WCS is not given it is calculated from the input WCSes.
       The first image is used as a reference, if no reference is given.
       This represents the virtual mosaic WCS.
    2. For each input observation/chip, an HSTWCS object is created
       and its footprint on the sky is calculated (using only the four corners).
    3. For each input observation the footprint is projected on the output
       tangent plane and the virtual WCS is recorded in the header.
    """
    wcsobjects = readWCS(fnames, ext, extname)
    if outwcs is not None:
        outwcs = outwcs.deepcopy()
    else:
        if ref_wcs is not None:
            outwcs = utils.output_wcs(wcsobjects, ref_wcs=ref_wcs, undistort=undistort)
        else:
            outwcs = utils.output_wcs(wcsobjects, undistort=undistort)
    if plot:
        outc = np.array([[0., 0], [outwcs.pixel_shape[0], 0],
                         [outwcs.pixel_shape[0], outwcs.pixel_shape[1]],
                         [0, outwcs.pixel_shape[1]], [0, 0]])
        plt.plot(outc[:, 0], outc[:, 1])
    for wobj in wcsobjects:
        outcorners = outwcs.wcs_world2pix(wobj.calc_footprint(), 1)
        if plot:
            plt.plot(outcorners[:, 0], outcorners[:, 1])
        objwcs = outwcs.deepcopy()
        objwcs.wcs.crpix = objwcs.wcs.crpix - (outcorners[0])
        updatehdr(wobj.filename, objwcs, wkey=wkey, wcsname=wname, ext=wobj.extname,
                  clobber=clobber)
    return outwcs


def updatehdr(fname, wcsobj, wkey, wcsname, ext=1, clobber=False):
    hdr = fits.getheader(fname, ext=ext)
    all_keys = list(string.ascii_uppercase)
    if wkey.upper() not in all_keys:
        raise KeyError("wkey must be one character: A-Z")
    if wkey not in altwcs.available_wcskeys(hdr):
        if not clobber:
            raise ValueError("wkey {0} is already in use."
                             "Use clobber=True to overwrite it or"
                             "specify a different key.".format(wkey))
        else:
            altwcs.deleteWCS(fname, ext=ext, wcskey='V')
    f = fits.open(fname, mode='update')

    hwcs = wcs2header(wcsobj)
    wcsnamekey = 'WCSNAME' + wkey
    f[ext].header[wcsnamekey] = wcsname
    for k in hwcs:
        f[ext].header[k[: 7] + wkey] = hwcs[k]

    f.close()


def wcs2header(wcsobj):
    h = wcsobj.to_header()

    if wcsobj.wcs.has_cd():
        altwcs.pc2cd(h)
    h['CTYPE1'] = 'RA---TAN'
    h['CTYPE2'] = 'DEC--TAN'
    norient = np.rad2deg(np.arctan2(h['CD1_2'], h['CD2_2']))
    # okey = 'ORIENT%s' % wkey
    okey = 'ORIENT'
    h[okey] = norient
    return h


def readWCS(input, exts=None, extname=None):
    if isinstance(input, str):
        if input[0] == '@':
            # input is an @ file
            filelist = irafglob.irafglob(input)
        else:
            try:
                filelist, output = parseinput.parseinput(input)
            except IOError: raise
    elif isinstance(input, list):
        if isinstance(input[0], wcsutil.HSTWCS):
            # a list of HSTWCS objects
            return input
        else:
            filelist = input[:]
    wcso = []
    fomited = []
    # figure out which FITS extension(s) to use
    if exts is None and extname is None:
        # Assume it's simple FITS and the data is in the primary HDU
        for f in filelist:
            try:
                wcso = wcsutil.HSTWCS(f)
            except AttributeError:
                fomited.append(f)
                continue
    elif exts is not None and validateExt(exts):
        exts = [exts]
        for f in filelist:
            try:
                wcso.extend([wcsutil.HSTWCS(f, ext=e) for e in exts])
            except KeyError:
                fomited.append(f)
                continue
    elif extname is not None:
        for f in filelist:
            fobj = fits.open(f)
            for i in range(len(fobj)):
                try:
                    ename = fobj[i].header['EXTNAME']
                except KeyError:
                    continue
                if ename.lower() == extname.lower():
                    wcso.append(wcsutil.HSTWCS(f, ext=i))
                else:
                    continue
            fobj.close()
    if fomited != []:
        print("These files were skipped:")
        for f in fomited:
            print(f)
    return wcso


def validateExt(ext):
    if not isinstance(ext, int) and not isinstance(ext, tuple) \
       and not isinstance(ext, list):
        print("Ext must be integer, tuple, a list of int extension numbers, \
        or a list of tuples representing a fits extension, for example ('sci', 1).")
        return False
    else:
        return True




import numpy as np
import logging
import time
from math import sin, sqrt, pow, cos, asin, atan2, pi

from stsci.tools import fileutil
from . import utils

logger = logging.getLogger(__name__)


class MakeWCS(object):
    """
    Recompute basic WCS keywords based on PA_V3 and distortion model.

    Notes
    -----
    - Compute the reference chip WCS:

        -- CRVAL: transform the model XREF/YREF to the sky
        -- PA_V3 is calculated at the target position and adjusted
           for each chip orientation
        -- CD: PA_V3 and the model scale are used to cnstruct a CD matrix

    - Compute the second chip WCS:

        -- CRVAL: - the distance between the zero points of the two
                  chip models on the sky
        -- CD matrix: first order coefficients are added to the components
            of this distance and transfered on the sky. The difference
            between CRVAL and these vectors is the new CD matrix for each chip.
        -- CRPIX: chip's model zero point in pixel space (XREF/YREF)

    - Time dependent distortion correction is applied to both chips' V2/V3 values.

    """
    tdd_xyref = {1: [2048, 3072], 2: [2048, 1024]}

    def updateWCS(cls, ext_wcs, ref_wcs):
        """
        recomputes the basic WCS kw
        """
        logger.info("\n\tStarting MakeWCS: %s" % time.asctime())
        if not ext_wcs.idcmodel:
            logger.info("IDC model not found, turning off Makewcs")
            return {}
        ltvoff, offshift = cls.getOffsets(ext_wcs)

        v23_corr = cls.zero_point_corr(ext_wcs)
        rv23_corr = cls.zero_point_corr(ref_wcs)

        cls.uprefwcs(ext_wcs, ref_wcs, rv23_corr, ltvoff, offshift)
        cls.upextwcs(ext_wcs, ref_wcs, v23_corr, rv23_corr, ltvoff, offshift)

        kw2update = {'CD1_1': ext_wcs.wcs.cd[0, 0],
                     'CD1_2': ext_wcs.wcs.cd[0, 1],
                     'CD2_1': ext_wcs.wcs.cd[1, 0],
                     'CD2_2': ext_wcs.wcs.cd[1, 1],
                     'CRVAL1': ext_wcs.wcs.crval[0],
                     'CRVAL2': ext_wcs.wcs.crval[1],
                     'CRPIX1': ext_wcs.wcs.crpix[0],
                     'CRPIX2': ext_wcs.wcs.crpix[1],
                     'IDCTAB': ext_wcs.idctab,
                     'OCX10': ext_wcs.idcmodel.cx[1, 0],
                     'OCX11': ext_wcs.idcmodel.cx[1, 1],
                     'OCY10': ext_wcs.idcmodel.cy[1, 0],
                     'OCY11': ext_wcs.idcmodel.cy[1, 1]
                     }
        return kw2update

    updateWCS = classmethod(updateWCS)

    def upextwcs(cls, ext_wcs, ref_wcs, v23_corr, rv23_corr, loff, offsh):
        """
        updates an extension wcs
        """
        ltvoffx, ltvoffy = loff[0], loff[1]
        offshiftx, offshifty = offsh[0], offsh[1]
        ltv1 = ext_wcs.ltv1
        ltv2 = ext_wcs.ltv2
        if ltv1 != 0. or ltv2 != 0.:
            offsetx = ext_wcs.wcs.crpix[0] - ltv1 - ext_wcs.idcmodel.refpix['XREF']
            offsety = ext_wcs.wcs.crpix[1] - ltv2 - ext_wcs.idcmodel.refpix['YREF']
            ext_wcs.idcmodel.shift(offsetx, offsety)

        fx, fy = ext_wcs.idcmodel.cx, ext_wcs.idcmodel.cy

        ext_wcs.setPscale()
        tddscale = (ref_wcs.pscale / fx[1, 1])
        v2 = ext_wcs.idcmodel.refpix['V2REF'] + v23_corr[0, 0] * tddscale
        v3 = ext_wcs.idcmodel.refpix['V3REF'] - v23_corr[1, 0] * tddscale
        v2ref = ref_wcs.idcmodel.refpix['V2REF'] + rv23_corr[0, 0] * tddscale
        v3ref = ref_wcs.idcmodel.refpix['V3REF'] - rv23_corr[1, 0] * tddscale

        R_scale = ref_wcs.idcmodel.refpix['PSCALE'] / 3600.0
        off = sqrt((v2 - v2ref) ** 2 + (v3 - v3ref) ** 2) / (R_scale * 3600.0)

        if v3 == v3ref:
            theta = 0.0
        else:
            theta = atan2(ext_wcs.parity[0][0] * (v2 - v2ref),
                          ext_wcs.parity[1][1] * (v3 - v3ref))

        if ref_wcs.idcmodel.refpix['THETA']: theta += ref_wcs.idcmodel.refpix['THETA'] * pi / 180.0

        dX = (off * sin(theta)) + offshiftx
        dY = (off * cos(theta)) + offshifty

        px = np.array([[dX, dY]])
        newcrval = ref_wcs.wcs.p2s(px, 1)['world'][0]
        newcrpix = np.array([ext_wcs.idcmodel.refpix['XREF'] + ltvoffx,
                             ext_wcs.idcmodel.refpix['YREF'] + ltvoffy])
        ext_wcs.wcs.crval = newcrval
        ext_wcs.wcs.crpix = newcrpix
        ext_wcs.wcs.set()

        # Create a small vector, in reference image pixel scale
        delmat = np.array([[fx[1, 1], fy[1, 1]],
                           [fx[1, 0], fy[1, 0]]]) / R_scale / 3600.

        # Account for subarray offset
        # Angle of chip relative to chip
        if ext_wcs.idcmodel.refpix['THETA']:
            dtheta = ext_wcs.idcmodel.refpix['THETA'] - ref_wcs.idcmodel.refpix['THETA']
        else:
            dtheta = 0.0

        rrmat = fileutil.buildRotMatrix(dtheta)
        # Rotate the vectors
        dxy = np.dot(delmat, rrmat)
        wc = ref_wcs.wcs.p2s((px + dxy), 1)['world']

        # Calculate the new CDs and convert to degrees
        cd11 = utils.diff_angles(wc[0, 0], newcrval[0]) * cos(newcrval[1] * pi / 180.0)
        cd12 = utils.diff_angles(wc[1, 0], newcrval[0]) * cos(newcrval[1] * pi / 180.0)
        cd21 = utils.diff_angles(wc[0, 1], newcrval[1])
        cd22 = utils.diff_angles(wc[1, 1], newcrval[1])
        cd = np.array([[cd11, cd12], [cd21, cd22]])
        ext_wcs.wcs.cd = cd
        ext_wcs.wcs.set()

    upextwcs = classmethod(upextwcs)

    def uprefwcs(cls, ext_wcs, ref_wcs, rv23_corr_tdd, loff, offsh):
        """
        Updates the reference chip
        """
        ltvoffx, ltvoffy = loff[0], loff[1]
        ltv1 = ref_wcs.ltv1
        ltv2 = ref_wcs.ltv2
        if ref_wcs.ltv1 != 0. or ref_wcs.ltv2 != 0.:
            offsetx = ref_wcs.wcs.crpix[0] - ltv1 - ref_wcs.idcmodel.refpix['XREF']
            offsety = ref_wcs.wcs.crpix[1] - ltv2 - ref_wcs.idcmodel.refpix['YREF']
            ref_wcs.idcmodel.shift(offsetx, offsety)

        # rfx, rfy = ref_wcs.idcmodel.cx, ref_wcs.idcmodel.cy

        # offshift = offsh
        dec = ref_wcs.wcs.crval[1]
        tddscale = (ref_wcs.pscale / ref_wcs.idcmodel.cx[1, 1])
        rv23 = [ref_wcs.idcmodel.refpix['V2REF'] + (rv23_corr_tdd[0, 0] * tddscale),
                ref_wcs.idcmodel.refpix['V3REF'] - (rv23_corr_tdd[1, 0] * tddscale)]
        # Get an approximate reference position on the sky
        rref = np.array([[ref_wcs.idcmodel.refpix['XREF'] + ltvoffx,
                          ref_wcs.idcmodel.refpix['YREF'] + ltvoffy]])

        crval = ref_wcs.wcs.p2s(rref, 1)['world'][0]
        # Convert the PA_V3 orientation to the orientation at the aperture
        # This is for the reference chip only - we use this for the
        # reference tangent plane definition
        # It has the same orientation as the reference chip
        pv = troll(ext_wcs.pav3, dec, rv23[0], rv23[1])
        # Add the chip rotation angle
        if ref_wcs.idcmodel.refpix['THETA']:
            pv += ref_wcs.idcmodel.refpix['THETA']

        # Set values for the rest of the reference WCS
        ref_wcs.wcs.crval = crval
        ref_wcs.wcs.crpix = np.array([0.0, 0.0]) + offsh
        parity = ref_wcs.parity
        R_scale = ref_wcs.idcmodel.refpix['PSCALE'] / 3600.0
        cd11 = parity[0][0] * cos(pv * pi / 180.0) * R_scale
        cd12 = parity[0][0] * -sin(pv * pi / 180.0) * R_scale
        cd21 = parity[1][1] * sin(pv * pi / 180.0) * R_scale
        cd22 = parity[1][1] * cos(pv * pi / 180.0) * R_scale

        rcd = np.array([[cd11, cd12], [cd21, cd22]])
        ref_wcs.wcs.cd = rcd
        ref_wcs.wcs.set()

    uprefwcs = classmethod(uprefwcs)

    def zero_point_corr(cls, hwcs):
        if hwcs.idcmodel.refpix['skew_coeffs'] is not None and 'TDD_CY_BETA' in hwcs.idcmodel.refpix['skew_coeffs']:
            v23_corr = np.array([[0.], [0.]])
            return v23_corr
        else:
            try:
                alpha = hwcs.idcmodel.refpix['TDDALPHA']
                beta = hwcs.idcmodel.refpix['TDDBETA']
            except KeyError:
                alpha = 0.0
                beta = 0.0
                v23_corr = np.array([[0.], [0.]])
                logger.debug("TDD Zero point correction for chip"
                             "{0} defaulted to: {1}".format(hwcs.chip, v23_corr))
                return v23_corr

        tdd = np.array([[beta, alpha], [alpha, -beta]])
        mrotp = fileutil.buildRotMatrix(2.234529) / 2048.
        xy0 = np.array([[cls.tdd_xyref[hwcs.chip][0] - 2048.],
                        [cls.tdd_xyref[hwcs.chip][1] - 2048.]])
        v23_corr = np.dot(mrotp, np.dot(tdd, xy0)) * 0.05
        logger.debug("TDD Zero point correction for chip {0}: {1}".format(hwcs.chip, v23_corr))
        return v23_corr

    zero_point_corr = classmethod(zero_point_corr)

    def getOffsets(cls, ext_wcs):
        ltv1 = ext_wcs.ltv1
        ltv2 = ext_wcs.ltv2

        offsetx = ext_wcs.wcs.crpix[0] - ltv1 - ext_wcs.idcmodel.refpix['XREF']
        offsety = ext_wcs.wcs.crpix[1] - ltv2 - ext_wcs.idcmodel.refpix['YREF']

        shiftx = ext_wcs.idcmodel.refpix['XREF'] + ltv1
        shifty = ext_wcs.idcmodel.refpix['YREF'] + ltv2
        if ltv1 != 0. or ltv2 != 0.:
            ltvoffx = ltv1 + offsetx
            ltvoffy = ltv2 + offsety
            offshiftx = offsetx + shiftx
            offshifty = offsety + shifty
        else:
            ltvoffx = 0.
            ltvoffy = 0.
            offshiftx = 0.
            offshifty = 0.

        ltvoff = np.array([ltvoffx, ltvoffy])
        offshift = np.array([offshiftx, offshifty])
        return ltvoff, offshift

    getOffsets = classmethod(getOffsets)


def troll(roll, dec, v2, v3):
    """ Computes the roll angle at the target position based on:
        the roll angle at the V1 axis(roll),
        the dec of the target(dec), and
        the V2/V3 position of the aperture (v2,v3) in arcseconds.

        Based on algorithm provided by Colin Cox and used in
        Generic Conversion at STScI.
    """
    # Convert all angles to radians
    _roll = np.deg2rad(roll)
    _dec = np.deg2rad(dec)
    _v2 = np.deg2rad(v2 / 3600.)
    _v3 = np.deg2rad(v3 / 3600.)

    # compute components
    sin_rho = sqrt((pow(sin(_v2), 2) + pow(sin(_v3), 2)) - (pow(sin(_v2), 2) * pow(sin(_v3), 2)))
    rho = asin(sin_rho)
    beta = asin(sin(_v3) / sin_rho)
    if _v2 < 0: beta = pi - beta
    gamma = asin(sin(_v2) / sin_rho)
    if _v3 < 0: gamma = pi - gamma
    A = pi / 2. + _roll - beta
    B = atan2(sin(A) * cos(_dec), (sin(_dec) * sin_rho - cos(_dec) * cos(rho) * cos(A)))

    # compute final value
    troll = np.rad2deg(pi - (gamma + B))

    return troll



import numpy as np
from astropy.io import fits
from astropy import wcs as pywcs


def sip2idc(wcs):
    """
    Converts SIP style coefficients to IDCTAB coefficients.

    Parameters
    ----------
    wcs : `astropy.io.fits.Header` or `astropy.wcs.WCS` object
    """
    if isinstance(wcs, fits.Header):
        ocx10 = wcs.get('OCX10', None)
        ocx11 = wcs.get('OCX11', None)
        ocy10 = wcs.get('OCY10', None)
        ocy11 = wcs.get('OCY11', None)
        order = wcs.get('A_ORDER', None)
        sipa, sipb = _read_sip_kw(wcs)
        if None in [ocx10, ocx11, ocy10, ocy11, sipa, sipb]:
            print('Cannot convert SIP to IDC coefficients.\n')
            return None, None
    elif isinstance(wcs, pywcs.WCS):
        try:
            ocx10 = wcs.ocx10
            ocx11 = wcs.ocx11
            ocy10 = wcs.ocy10
            ocy11 = wcs.ocy11
        except AttributeError:
            print('First order IDCTAB coefficients are not available.\n')
            print('Cannot convert SIP to IDC coefficients.\n')
            return None, None
        try:
            sipa = wcs.sip.a
            sipb = wcs.sip.b
        except AttributeError:
            print('SIP coefficients are not available.')
            print('Cannot convert SIP to IDC coefficients.\n')
            return None, None
        try:
            order = wcs.sip.a_order
        except AttributeError:
            print('SIP model order unknown, exiting ...\n')
            return None, None

    else:
        print('Input to sip2idc must be a PyFITS header or a wcsutil.HSTWCS object\n')
        return

    if None in [ocx10, ocx11, ocy10, ocy11]:
        print('First order IDC coefficients not found, exiting ...\n')
        return None, None
    idc_coeff = np.array([[ocx11, ocx10], [ocy11, ocy10]])
    cx = np.zeros((order + 1, order + 1), dtype=np.double)
    cy = np.zeros((order + 1, order + 1), dtype=np.double)
    for n in range(order + 1):
        for m in range(order + 1):
            if n >= m and n >= 2:
                sipval = np.array([[sipa[m, n - m]], [sipb[m, n - m]]])
                idcval = np.dot(idc_coeff, sipval)
                cx[n, m] = idcval[0]
                cy[n, m] = idcval[1]

    cx[1, 0] = ocx10
    cx[1, 1] = ocx11
    cy[1, 0] = ocy10
    cy[1, 1] = ocy11

    return cx, cy


def _read_sip_kw(header):
    """
    Reads SIP header keywords and returns an array of coefficients.

    If no SIP header keywords are found, None is returned.
    """
    if "A_ORDER" in header:
        if "B_ORDER" not in header:
            raise ValueError(
                "A_ORDER provided without corresponding B_ORDER "
                "keyword for SIP distortion")

        m = int(header["A_ORDER"])
        a = np.zeros((m + 1, m + 1), np.double)
        for i in range(m + 1):
            for j in range(m - i + 1):
                a[i, j] = header.get("A_%d_%d" % (i, j), 0.0)

        m = int(header["B_ORDER"])
        b = np.zeros((m + 1, m + 1), np.double)
        for i in range(m + 1):
            for j in range(m - i + 1):
                b[i, j] = header.get("B_%d_%d" % (i, j), 0.0)
    elif "B_ORDER" in header:
        raise ValueError(
            "B_ORDER provided without corresponding A_ORDER "
            "keyword for SIP distortion")
    else:
        a = None
        b = None

    return a, b


"""
def idc2sip(wcsobj, idctab = None):
    if isinstance(wcs,pywcs.WCS):
        try:
            cx10 = wcsobj.ocx10
            cx11 = wcsobj.cx11
            cy10 = wcsobj.cy10
            cy11 = wcsobj.cy11
        except AttributeError:
            print
        try:
            order = wcs.sip.a_order
        except AttributeError:
            print 'SIP model order unknown, exiting ...\n'
            return
    else:
        print 'Input to sip2idc must be a PyFITS header or a wcsutil.HSTWCS object\n'
        return

    if None in [ocx10, ocx11, ocy10, ocy11]:
        print 'First order IDC coefficients not found, exiting ...\n'
        return
    idc_coeff = np.array([[wcsobj.cx11, wcsobj.cx10], [wcsobj.cy11, wcsobj.cy10]])
    cx = numpy.zeros((order+1,order+1), dtype=numpy.double)
    cy = numpy.zeros((order+1,order+1), dtype=numpy.double)
    for n in range(order+1):
        for m in range(order+1):
            if n >= m and n>=2:
                sipval = numpy.array([[wcsobj.sip.a[n,m]],[wcsobj.sip.b[n,m]]])
                idcval = numpy.dot(idc_coeff, sipval)
                cx[m,n-m] = idcval[0]
                cy[m,n-m] = idcval[1]

    return cx, cy
"""



from astropy.io import fits
from stsci.tools import irafglob, fileutil, parseinput


def parseSingleInput(f=None, ext=None):
    if isinstance(f, str):
        # create an HSTWCS object from a filename
        if ext is not None:
            filename = f
            if isinstance(ext, tuple):
                if ext[0] == '':
                    extnum = ext[1]  # handle ext=('',1)
                else:
                    extnum = ext
            else:
                extnum = int(ext)
        elif ext is None:
            filename, ext = fileutil.parseFilename(f)
            ext = fileutil.parseExtn(ext)
            if ext[0] == '':
                extnum = int(ext[1])  # handle ext=('',extnum)
            else:
                extnum = ext
        phdu = fits.open(filename)
        hdr0 = phdu[0].header
        try:
            ehdr = phdu[extnum].header
        except (IndexError, KeyError) as e:
            raise e.__class__('Unable to get extension %s.' % extnum)

    elif isinstance(f, fits.HDUList):
        phdu = f
        if ext is None:
            extnum = 0
        else:
            extnum = ext
        ehdr = f[extnum].header
        hdr0 = f[0].header
        filename = hdr0.get('FILENAME', "")

    else:
        raise ValueError('Input must be a file name string or a'
                         '`astropy.io.fits.HDUList` object')

    return filename, hdr0, ehdr, phdu


def parseMultipleInput(input):
    if isinstance(input, str):
        if input[0] == '@':
            # input is an @ file
            filelist = irafglob.irafglob(input)
        else:
            try:
                filelist, output = parseinput.parseinput(input)
            except IOError: raise
    elif isinstance(input, list):
        #if isinstance(input[0], HSTWCS):
            ## a list of HSTWCS objects
            #return input
        #else:
        filelist = input[:]
    return filelist



import numpy as np
from . import mutil
from .mutil import combin


#################
#
#
#               Geometry/Distortion Classes
#
#
#################

class GeometryModel:
    """
    Base class for Distortion model.
    There will be a separate class for each type of
    model/filetype used with drizzle, i.e., IDCModel and
    DrizzleModel.

    Each class will know how to apply the distortion to a
    single point and how to convert coefficients to an input table
    suitable for the drizzle task.

    Coefficients will be stored in CX,CY arrays.
    """
    #
    #
    #
    #
    #
    #
    #
    NORDER = 3

    def __init__(self):
        "  This will open the given file and determine its type and norder."

        #       Method to read in coefficients from given table and
        #       populate the n arrays 'cx' and 'cy'.
        #       This will be different for each type of input file,
        #       IDCTAB vs. drizzle table.

        # Set these up here for all sub-classes to use...
        # But, calculate norder and cx,cy arrays in detector specific classes.
        self.cx = None
        self.cy = None
        self.refpix = None
        self.norder = self.NORDER
        # Keep track of computed zero-point for distortion coeffs
        self.x0 = None
        self.y0 = None

        # default values for these attributes
        self.direction = 'forward'

        self.pscale = 1.0

    def shift(self, xs, ys):
        """
        Shift reference position of coefficients to new center
        where (xs,ys) = old-reference-position - subarray/image center.
        This will support creating coeffs files for drizzle which will
        be applied relative to the center of the image, rather than relative
        to the reference position of the chip.
        """

        _cxs = np.zeros(shape=self.cx.shape, dtype=self.cx.dtype)
        _cys = np.zeros(shape=self.cy.shape, dtype=self.cy.dtype)
        _k = self.norder + 1
        # loop over each input coefficient
        for m in range(_k):
            for n in range(_k):
                if m >= n:
                    # For this coefficient, shift by xs/ys.
                    _ilist = list(range(m, _k))
                    # sum from m to k
                    for i in _ilist:
                        _jlist = list(range(n, i - (m - n) + 1))
                        # sum from n to i-(m-n)
                        for j in _jlist:
                            _cxs[m, n] += self.cx[i, j] * combin(j, n) * combin((i - j), (m - n)) * \
                                pow(xs, (j - n)) * pow(ys, ((i - j) - (m - n)))
                            _cys[m, n] += self.cy[i, j] * combin(j, n) * combin((i - j), (m - n)) * \
                                pow(xs, (j - n)) * pow(ys, ((i - j) - (m - n)))
        self.cx = _cxs.copy()
        self.cy = _cys.copy()

    def convert(self, tmpname, xref=None, yref=None, delta=True):
        """
         Open up an ASCII file, output coefficients in drizzle
          format after converting them as necessary.
        First, normalize these coefficients to what drizzle expects
        Normalize the coefficients by the MODEL/output plate scale.

        16-May-2002:
        Revised to work with higher order polynomials by John Blakeslee.
        27-June-2002:
            Added ability to shift coefficients to new center for support
                of subarrays.
        """
        cx = self.cx / self.pscale
        cy = self.cy / self.pscale
        x0 = self.refpix['XDELTA'] + cx[0, 0]
        y0 = self.refpix['YDELTA'] + cy[0, 0]
        xr = self.refpix['CHIP_XREF']
        yr = self.refpix['CHIP_YREF']

        self.x0 = x0
        self.y0 = y0

        # Now, write out the coefficients into an ASCII
        # file in 'drizzle' format.
        lines = []
        lines.append('# Polynomial distortion coefficients\n')
        lines.append('# Extracted from "%s" \n' % self.name)
        lines.append('refpix %f %f \n' % (xr, yr))
        if self.norder == 3:
            lines.append('cubic\n')
        elif self.norder == 4:
            lines.append('quartic\n')
        elif self.norder == 5:
            lines.append('quintic\n')
        else:
            raise ValueError("Drizzle cannot handle poly distortions of order %d" % self.norder)

        str = '%16.8f %16.8g %16.8g %16.8g %16.8g \n' % (x0, cx[1, 1], cx[1, 0], cx[2, 2], cx[2, 1])
        lines.append(str)
        str = '%16.8g %16.8g %16.8g %16.8g %16.8g \n' % (cx[2, 0], cx[3, 3], cx[3, 2], cx[3, 1], cx[3, 0])
        lines.append(str)
        if self.norder > 3:
            str = '%16.8g %16.8g %16.8g %16.8g %16.8g \n' % (cx[4, 4], cx[4, 3], cx[4, 2], cx[4, 1], cx[4, 0])
            lines.append(str)
        if self.norder > 4:
            str = '%16.8g %16.8g %16.8g %16.8g %16.8g %16.8g \n' % (cx[5, 5], cx[5, 4], cx[5, 3], cx[5, 2], cx[5, 1], cx[5, 0])
            lines.append(str)
        lines.append("\n")

        str = '%16.8f %16.8g %16.8g %16.8g %16.8g \n' % (y0, cy[1, 1], cy[1, 0], cy[2, 2], cy[2, 1])
        lines.append(str)
        str = '%16.8g %16.8g %16.8g %16.8g %16.8g \n' % (cy[2, 0], cy[3, 3], cy[3, 2], cy[3, 1],
                                                         cy[3, 0])
        lines.append(str)
        if self.norder > 3:
            str = '%16.8g %16.8g %16.8g %16.8g %16.8g \n' % (cy[4, 4], cy[4, 3], cy[4, 2],
                                                             cy[4, 1], cy[4, 0])
            lines.append(str)
        if self.norder > 4:
            str = '%16.8g %16.8g %16.8g %16.8g %16.8g %16.8g \n' % (cy[5, 5], cy[5, 4], cy[5, 3],
                                                                    cy[5, 2], cy[5, 1], cy[5, 0])
            lines.append(str)

        output = open(tmpname, 'w')
        output.writelines(lines)
        output.close()

    def apply(self, pixpos, scale=1.0, order=None):
        """
         Apply coefficients to a pixel position or a list of positions.
          This should be the same for all coefficients tables.
        Return the geometrically-adjusted position
        in arcseconds from the reference position as a tuple (x,y).

        Compute delta from reference position
        """

        """
        scale actually is a ratio of pscale/self.model.pscale
        what is pscale?
        """
        if self.cx is None:
            return pixpos[:, 0], pixpos[:, 1]

        if order is None:
            order = self.norder

        # Apply in the same way that 'drizzle' would...
        _cx = self.cx / (self.pscale * scale)
        _cy = self.cy / (self.pscale * scale)
        _convert = False
        _p = pixpos

        # Do NOT include any zero-point terms in CX,CY here
        # as they should not be scaled by plate-scale like rest
        # of coeffs...  This makes the computations consistent
        # with 'drizzle'.  WJH 17-Feb-2004
        _cx[0, 0] = 0.
        _cy[0, 0] = 0.

        if isinstance(_p, list) or isinstance(_p, tuple):
            _p = np.array(_p, dtype=np.float64)
            _convert = True

        dxy = _p - (self.refpix['XREF'], self.refpix['YREF'])
        # Apply coefficients from distortion model here...
        c = _p * 0.
        for i in range(order + 1):
            for j in range(i + 1):
                c[:, 0] = c[:, 0] + _cx[i][j] * pow(dxy[:, 0], j) * pow(dxy[:, 1], (i - j))
                c[:, 1] = c[:, 1] + _cy[i][j] * pow(dxy[:, 0], j) * pow(dxy[:, 1], (i - j))
        xc = c[:, 0]
        yc = c[:, 1]

        # Convert results back to same form as original input
        if _convert:
            xc = xc.tolist()
            yc = yc.tolist()
            # If a single tuple was input, return just a single tuple
            if len(xc) == 1:
                xc = xc[0]
                yc = yc[0]

        return xc, yc

    def setPScaleCoeffs(self, pscale):
        self.cx[1, 1] = pscale
        self.cy[1, 0] = pscale

        self.refpix['PSCALE'] = pscale
        self.pscale = pscale


class IDCModel(GeometryModel):
    """
    This class will open the IDCTAB, select proper row based on
    chip/direction and populate cx,cy arrays.
    We also need to read in SCALE, XCOM,YCOM, XREF,YREF as well.
    """
    def __init__(self, idcfile, date=None, chip=1, direction='forward',
                 filter1='CLEAR1', filter2='CLEAR2', offtab=None, binned=1):
        GeometryModel.__init__(self)
        #
        # Norder must be derived from the coeffs file itself,
        # then the arrays can be setup. Thus, it needs to be
        # done in the sub-class, not in the base class.
        # Read in table.
        # Populate cx,cy,scale, and other variables here.
        #
        self.name = idcfile
        self.cx, self.cy, self.refpix, self.norder = mutil.readIDCtab(idcfile,
                                                                      chip=chip, direction=direction, filter1=filter1, filter2=filter2,
                                                                      date=date, offtab=offtab)

        if 'empty_model' in self.refpix and self.refpix['empty_model']:
            pass
        else:
            self.refpix['PSCALE'] = self.refpix['PSCALE'] * binned
            self.cx = self.cx * binned
            self.cy = self.cy * binned
            self.refpix['XREF'] = self.refpix['XREF'] / binned
            self.refpix['YREF'] = self.refpix['YREF'] / binned
            self.refpix['XSIZE'] = self.refpix['XSIZE'] / binned
            self.refpix['YSIZE'] = self.refpix['YSIZE'] / binned

        self.pscale = self.refpix['PSCALE']


class WCSModel(GeometryModel):
    """
    This class sets up a distortion model based on coefficients
    found in the image header.
    """
    def __init__(self, header, rootname):
        GeometryModel.__init__(self)

        if 'rootname' in header:
            self.name = header['rootname']
        else:
            self.name = rootname
        # Initialize all necessary distortion arrays with
        # default model...
        # self.cx,self.cy,self.refpix,self.order = mutil.defaultModel()

        # Read in values from header, and update distortion arrays.
        self.cx, self.cy, self.refpix, self.norder = mutil.readWCSCoeffs(header)

        self.pscale = self.refpix['PSCALE']


class DrizzleModel(GeometryModel):
    """
    This class will read in an ASCII Cubic
    drizzle coeffs file and populate the cx,cy arrays.
    """

    def __init__(self, idcfile, scale=None):
        GeometryModel.__init__(self)
        #
        # We now need to read in the file, populate cx,cy, and
        # other variables as necessary.
        #
        self.name = idcfile
        self.cx, self.cy, self.refpix, self.norder = mutil.readCubicTable(idcfile)

        # scale is the ratio wcs.pscale/model.pscale.
        # model.pscale for WFPC2 is passed from REFDATA.
        # This is needed for WFPC2 binned data.

        if scale is not None:
            self.pscale = scale
        else:
            self.pscale = self.refpix['PSCALE']


class TraugerModel(GeometryModel):
    """
    This class will read in the ASCII Trauger coeffs
    file, convert them to SIAF coefficients, then populate
    the cx,cy arrays.
    """
    NORDER = 3

    def __init__(self, idcfile, lam):
        GeometryModel.__init__(self)
        self.name = idcfile
        self.cx, self.cy, self.refpix, self.norder = mutil.readTraugerTable(idcfile, lam)
        self.pscale = self.refpix['PSCALE']



import string

import numpy as np
from astropy import wcs as pywcs
from astropy.io import fits
from stsci.tools import fileutil as fu

from astropy import log
default_log_level = log.getEffectiveLevel()


__all__ = ["archiveWCS", "available_wcskeys", "convertAltWCS", "deleteWCS", "next_wcskey",
           "pc2cd", "readAltWCS", "restoreWCS", "wcskeys", "wcsnames"]


altwcskw = ['WCSAXES', 'CRVAL', 'CRPIX', 'PC', 'CDELT', 'CD', 'CTYPE', 'CUNIT',
            'PV', 'PS']
altwcskw_extra = ['LATPOLE', 'LONPOLE', 'RESTWAV', 'RESTFRQ']

# file operations


def archiveWCS(fname, ext, wcskey=" ", wcsname=" ", reusekey=False):
    """
    Copy the primary WCS to the header as an alternate WCS
    with wcskey and name WCSNAME. It loops over all extensions in 'ext'

    Parameters
    ----------
    fname :  string or `astropy.io.fits.HDUList`
        file name or a file object
    ext :    int, tuple, str, or list of integers or tuples (e.g.('sci',1))
        fits extensions to work with
        If a string is provided, it should specify the EXTNAME of extensions
        with WCSs to be archived
    wcskey : string "A"-"Z" or " "
        if " ": get next available key if wcsname is also " " or try
        to get a key from WCSNAME value
    wcsname : string
        Name of alternate WCS description
    reusekey : boolean
        if True - overwrites a WCS with the same key

    Examples
    --------
    Copy the primary WCS of an in memory headrlet object to an
    alternate WCS with key 'T'

    >>> hlet=headerlet.createHeaderlet('junk.fits', 'hdr1.fits')
    >>> altwcs.wcskeys(hlet[1].header)
    ['A']
    >>> altwcs.archiveWCS(hlet, ext=[('SIPWCS',1),('SIPWCS',2)], wcskey='T')
    >>> altwcs.wcskeys(hlet[1].header)
    ['A', 'T']


    See Also
    --------
    wcsutil.restoreWCS: Copy an alternate WCS to the primary WCS

    """

    if isinstance(fname, str):
        f = fits.open(fname, mode='update')
    else:
        f = fname

    if not _parpasscheck(f, ext, wcskey, wcsname):
        closefobj(fname, f)
        raise ValueError("Input parameters problem")

    # Interpret input 'ext' value to get list of extensions to process
    ext = _buildExtlist(f, ext)

    if not wcskey and not wcsname:
        raise KeyError("Either wcskey or wcsname should be specified")

    if wcsname.strip() == "":
        try:
            wcsname = readAltWCS(f, ext[0], wcskey=" ")['WCSNAME']
        except KeyError:
            pass
    wcsext = ext[0]
    if wcskey != " " and wcskey in wcskeys(f[wcsext].header) and not reusekey:
        closefobj(fname, f)
        raise KeyError("Wcskey %s is aready used. \
        Run archiveWCS() with reusekey=True to overwrite this alternate WCS. \
        Alternatively choose another wcskey with altwcs.available_wcskeys()." % wcskey)
    elif wcskey == " ":
        # wcsname exists, overwrite it if reuse is True or get the next key
        if wcsname.strip() in wcsnames(f[wcsext].header).values():
            if reusekey:
                # try getting the key from an existing WCS with WCSNAME
                wkey = getKeyFromName(f[wcsext].header, wcsname)
                wname = wcsname
                if wkey == ' ':
                    wkey = next_wcskey(f[wcsext].header)
                elif wkey is None:
                    closefobj(fname, f)
                    raise KeyError("Could not get a valid wcskey from wcsname %s" % wcsname)
            else:
                closefobj(fname, f)
                raise KeyError("Wcsname %s is aready used. \
                Run archiveWCS() with reusekey=True to overwrite this alternate WCS. \
                Alternatively choose another wcskey with altwcs.available_wcskeys() or\
                choose another wcsname." % wcsname)
        else:
            wkey = next_wcskey(f[wcsext].header)
            if wcsname.strip():
                wname = wcsname
            else:
                # determine which WCSNAME needs to be replicated in archived WCS
                wnames = wcsnames(f[wcsext].header)
                if 'O' in wnames: del wnames['O']  # we don't want OPUS/original
                if len(wnames) > 0:
                    if ' ' in wnames:
                        wname = wnames[' ']
                    else:
                        akeys = string.ascii_uppercase
                        wname = "DEFAULT"
                        for key in akeys[-1::]:
                            if key in wnames:
                                wname = wnames
                                break
                else:
                    wname = "DEFAULT"
    else:
        wkey = wcskey
        wname = wcsname
    log.setLevel('WARNING')
    for e in ext:
        hdr = _getheader(f, e)
        w = pywcs.WCS(hdr, f)
        hwcs = w.to_header()

        if hwcs is None:
            continue

        if w.sip is not None:
            for i in range(1, w.naxis + 1):
                hwcs['CTYPE{0}'.format(i)] = hwcs['CTYPE{0}'.format(i)] + '-SIP'

        if w.wcs.has_cd():
            hwcs = pc2cd(hwcs, key=" ")

        wcsnamekey = 'WCSNAME' + wkey
        f[e].header[wcsnamekey] = wname

        try:
            old_wcsname = hwcs.pop('WCSNAME')
        except:
            pass

        for k in hwcs.keys():
            key = k[: 7] + wkey
            f[e].header[key] = hwcs[k]
    log.setLevel(default_log_level)
    closefobj(fname, f)


def restore_from_to(f, fromext=None, toext=None, wcskey=" ", wcsname=" "):
    """
    Copy an alternate WCS from one extension as a primary WCS of another extension

    Reads in a WCS defined with wcskey and saves it as the primary WCS.
    Goes sequentially through the list of extensions in ext.
    Alternatively uses 'fromext' and 'toext'.


    Parameters
    ----------
    f:       string or `astropy.io.fits.HDUList`
             a file name or a file object
    fromext: string
             extname from which to read in the alternate WCS, for example 'SCI'
    toext:   string or python list
             extname or a list of extnames to which the WCS will be copied as
             primary, for example ['SCI', 'ERR', 'DQ']
    wcskey:  a charater
             "A"-"Z" - Used for one of 26 alternate WCS definitions.
             or " " - find a key from WCSNAMe value
    wcsname: string (optional)
             if given and wcskey is " ", will try to restore by WCSNAME value

    See Also
    --------
    archiveWCS - copy the primary WCS as an alternate WCS
    restoreWCS - Copy a WCS with key "WCSKEY" to the primary WCS

    """
    if isinstance(f, str):
        fobj = fits.open(f, mode='update')
    else:
        fobj = f

    if not _parpasscheck(fobj, ext=None, wcskey=wcskey, fromext=fromext, toext=toext):
        closefobj(f, fobj)
        raise ValueError("Input parameters problem")

    # Interpret input 'ext' value to get list of extensions to process
    # ext = _buildExtlist(fobj, ext)

    if isinstance(toext, str):
        toext = [toext]

    # the case of an HDUList object in memory without an associated file

    # if fobj.filename() is not None:
    #        name = fobj.filename()

    simplefits = fu.isFits(fobj)[1] is 'simple'
    if simplefits:
        wcskeyext = 0
    else:
        wcskeyext = 1

    if wcskey == " ":
        if wcsname.strip():
            wkey = getKeyFromName(fobj[wcskeyext].header, wcsname)
            if not wkey:
                closefobj(f, fobj)
                raise KeyError("Could not get a key from wcsname %s ." % wcsname)
    else:
        if wcskey not in wcskeys(fobj, ext=wcskeyext):
            print("Could not find alternate WCS with key %s in this file" % wcskey)
            closefobj(f, fobj)
            return
        wkey = wcskey

    countext = fu.countExtn(fobj, fromext)
    if not countext:
        raise KeyError("File does not have extension with extname %s", fromext)
    else:
        for i in range(1, countext + 1):
            for toe in toext:
                _restore(fobj, fromextnum=i, fromextnam=fromext, toextnum=i, toextnam=toe, ukey=wkey)

    if fobj.filename() is not None:
        # fobj.writeto(name)
        closefobj(f, fobj)


def restoreWCS(f, ext, wcskey=" ", wcsname=" "):
    """
    Copy a WCS with key "WCSKEY" to the primary WCS

    Reads in a WCS defined with wcskey and saves it as the primary WCS.
    Goes sequentially through the list of extensions in ext.
    Alternatively uses 'fromext' and 'toext'.


    Parameters
    ----------
    f : str or `astropy.io.fits.HDUList`
        file name or a file object
    ext : int, tuple, str, or list of integers or tuples (e.g.('sci',1))
        fits extensions to work with
        If a string is provided, it should specify the EXTNAME of extensions
        with WCSs to be archived
    wcskey : str
        "A"-"Z" - Used for one of 26 alternate WCS definitions.
        or " " - find a key from WCSNAMe value
    wcsname : str
        (optional) if given and wcskey is " ", will try to restore by WCSNAME value

    See Also
    --------
    archiveWCS - copy the primary WCS as an alternate WCS
    restore_from_to

    """
    if isinstance(f, str):
        fobj = fits.open(f, mode='update')
    else:
        fobj = f

    if not _parpasscheck(fobj, ext=ext, wcskey=wcskey):
        closefobj(f, fobj)
        raise ValueError("Input parameters problem")

    # Interpret input 'ext' value to get list of extensions to process
    ext = _buildExtlist(fobj, ext)

    # the case of an HDUList object in memory without an associated file

    simplefits = fu.isFits(fobj)[1] is 'simple'
    if simplefits:
        wcskeyext = 0
    else:
        wcskeyext = 1

    if wcskey == " ":
        if wcsname.strip():
            wcskey = getKeyFromName(fobj[wcskeyext].header, wcsname)
            if not wcskey:
                closefobj(f, fobj)
                raise KeyError("Could not get a key from wcsname %s ." % wcsname)

    for e in ext:
        if wcskey not in wcskeys(fobj, ext=e):
            continue
        else:
            _restore(fobj, wcskey, fromextnum=e, verbose=False)

    if fobj.filename() is not None:
        closefobj(f, fobj)


def deleteWCS(fname, ext, wcskey=" ", wcsname=" "):
    """
    Delete an alternate WCS defined with wcskey.
    If wcskey is " " try to get a key from WCSNAME.

    Parameters
    ----------
    fname : str or a `astropy.io.fits.HDUList`
    ext : int, tuple, str, or list of integers or tuples (e.g.('sci',1))
        fits extensions to work with
        If a string is provided, it should specify the EXTNAME of extensions
        with WCSs to be archived
    wcskey : str
        one of 'A'-'Z' or " "
    wcsname : str
        Name of alternate WCS description
    """
    if isinstance(fname, str):
        fobj = fits.open(fname, mode='update')
    else:
        fobj = fname

    if not _parpasscheck(fobj, ext, wcskey, wcsname):
        closefobj(fname, fobj)
        raise ValueError("Input parameters problem")

    # Interpret input 'ext' value to get list of extensions to process
    ext = _buildExtlist(fobj, ext)
    # Do not allow deleting the original WCS.
    if wcskey == 'O':
        print("Wcskey 'O' is reserved for the original WCS and should not be deleted.")
        closefobj(fname, fobj)
        return

    wcskeyext = ext[0]

    if not wcskeys and not wcsname:
        raise KeyError("Either wcskey or wcsname should be specified")

    if wcskey == " ":
        # try getting the key from WCSNAME
        wkey = getKeyFromName(fobj[wcskeyext].header, wcsname)
        if not wkey:
            closefobj(fname, fobj)
            raise KeyError("Could not get a key: wcsname '%s' not found in header." % wcsname)
    else:
        if wcskey not in wcskeys(fobj[wcskeyext].header):
            closefobj(fname, fobj)
            raise KeyError("Could not find alternate WCS with key %s in this file" % wcskey)
        wkey = wcskey

    prexts = []
    for i in ext:
        hdr = fobj[i].header
        hwcs = readAltWCS(fobj, i, wcskey=wkey)
        if hwcs is None:
            continue
        for k in hwcs[::-1]:
            try:
                del hdr[k]
            except KeyError:
                pass
        prexts.append(i)
    if prexts != []:
        print('Deleted all instances of WCS with key %s in extensions' % wkey, prexts)
    else:
        print("Did not find WCS with key %s in any of the extensions" % wkey)
    closefobj(fname, fobj)


def _buildExtlist(fobj, ext):
    """
    Utility function to interpret the provided value of 'ext' and return a list
    of 'valid' values which can then be used by the rest of the functions in
    this module.

    Parameters
    ----------
    fobj: HDUList
        file to be examined
    ext:    an int, a tuple, string, list of integers or tuples (e.g.('sci',1))
            fits extensions to work with
            If a string is provided, it should specify the EXTNAME of extensions
            with WCSs to be archived
    """
    if not isinstance(ext, list):
        if isinstance(ext, str):
            extstr = ext
            ext = []
            for extn in range(1, len(fobj)):
                if 'extname' in fobj[extn].header and fobj[extn].header['extname'] == extstr:
                    ext.append(extn)
        elif isinstance(ext, int) or isinstance(ext, tuple):
            ext = [ext]
        else:
            raise KeyError("Valid extensions in 'ext' parameter need to be specified.")
    return ext


def _restore(fobj, ukey, fromextnum,
             toextnum=None, fromextnam=None, toextnam=None, verbose=True):
    """
    fobj: string of HDUList
    ukey: string 'A'-'Z'
          wcs key
    fromextnum: int
                extver of extension from which to copy WCS
    fromextnam: string
                extname of extension from which to copy WCS
    toextnum: int
              extver of extension to which to copy WCS
    toextnam: string
              extname of extension to which to copy WCS
    """
    # create an extension tuple, e.g. ('SCI', 2)
    if fromextnam:
        fromextension = (fromextnam, fromextnum)
    else:
        fromextension = fromextnum
    if toextnum:
        if toextnam:
            toextension = (toextnam, toextnum)
        else:
            toextension = toextnum
    else:
        toextension = fromextension

    hdr = _getheader(fobj, fromextension)
    # keep a copy of the ctype because of the "-SIP" suffix.
    ctype = hdr['ctype*']

    w = pywcs.WCS(hdr, fobj, key=ukey)
    hwcs = w.to_header()
    if hwcs is None:
        return

    if w.wcs.has_cd():
        hwcs = pc2cd(hwcs, key=ukey)

    for i in range(1, w.naxis + 1):
        hwcs['CTYPE{0}{1}'.format(i, ukey)] = ctype['CTYPE{0}'.format(i)]

    for k in hwcs.keys():
        key = k[:-1]
        if key in fobj[toextension].header:
            fobj[toextension].header[key] = hwcs[k]
        else:
            continue

    if key == 'O' and 'TDDALPHA' in fobj[toextension].header:
        fobj[toextension].header['TDDALPHA'] = 0.0
        fobj[toextension].header['TDDBETA'] = 0.0
    if 'ORIENTAT' in fobj[toextension].header:
        norient = np.rad2deg(np.arctan2(hwcs['CD1_2' + '%s' % ukey], hwcs['CD2_2' + '%s' % ukey]))
        fobj[toextension].header['ORIENTAT'] = norient
    # Reset 2014 TDD keywords prior to computing new values (if any are computed)
    for kw in ['TDD_CYA', 'TDD_CYB', 'TDD_CXA', 'TDD_CXB']:
        if kw in fobj[toextension].header:
            fobj[toextension].header[kw] = 0.0


# header operations


def _check_headerpars(fobj, ext):
    if not isinstance(fobj, fits.Header) and not isinstance(fobj, fits.HDUList) \
            and not isinstance(fobj, str):
        raise ValueError("Expected a file name, a file object or a header\n")

    if not isinstance(fobj, fits.Header):
        if not isinstance(ext, int) and not isinstance(ext, tuple):
            raise ValueError("Expected ext to be a number or a tuple, e.g. ('SCI', 1)\n")


def _getheader(fobj, ext):
    if isinstance(fobj, str):
        hdr = fits.getheader(fobj, ext)
    elif isinstance(fobj, fits.Header):
        hdr = fobj
    else:
        hdr = fobj[ext].header
    return hdr


def readAltWCS(fobj, ext, wcskey=' ', verbose=False):
    """
    Reads in alternate primary WCS from specified extension.

    Parameters
    ----------
    fobj : str, `astropy.io.fits.HDUList`
        fits filename or fits file object
        containing alternate/primary WCS(s) to be converted
    wcskey : str
        [" ",A-Z]
        alternate/primary WCS key that will be replaced by the new key
    ext : int
        fits extension number

    Returns
    -------
    hdr: fits.Header
        header object with ONLY the keywords for specified alternate WCS
    """
    if isinstance(fobj, str):
        fobj = fits.open(fobj)

    hdr = _getheader(fobj, ext)
    try:
        nwcs = pywcs.WCS(hdr, fobj=fobj, key=wcskey)
    except KeyError:
        if verbose:
            print('readAltWCS: Could not read WCS with key %s' % wcskey)
            print('            Skipping %s[%s]' % (fobj.filename(), str(ext)))
        return None
    hwcs = nwcs.to_header()

    if nwcs.wcs.has_cd():
        hwcs = pc2cd(hwcs, key=wcskey)

    return hwcs


def convertAltWCS(fobj, ext, oldkey=" ", newkey=' '):
    """
    Translates the alternate/primary WCS with one key to an alternate/primary WCS with
    another key.

    Parameters
    ----------
    fobj : str, `astropy.io.fits.HDUList`, or `astropy.io.fits.Header`
        fits filename, fits file object or fits header
        containing alternate/primary WCS(s) to be converted
    ext : int
        extension number
    oldkey : str
        [" ",A-Z]
        alternate/primary WCS key that will be replaced by the new key
    newkey : str
        [" ",A-Z]
        new alternate/primary WCS key

    Returns
    -------
    hdr: `astropy.io.fits.Header`
        header object with keywords renamed from oldkey to newkey
    """
    hdr = readAltWCS(fobj, ext, wcskey=oldkey)
    if hdr is None:
        return None
    # Converting WCS to new key
    for card in hdr:
        if oldkey == ' ' or oldkey == '':
            cname = card
        else:
            cname = card.rstrip(oldkey)
        hdr.rename_keyword(card, cname + newkey, force=True)

    return hdr


def wcskeys(fobj, ext=None):
    """
    Returns a list of characters used in the header for alternate
    WCS description with WCSNAME keyword

    Parameters
    ----------
    fobj : str, `astropy.io.fits.HDUList` or `astropy.io.fits.Header`
          fits file name, fits file object or fits header
    ext : int or None
         extension number
         if None, fobj must be a header
    """
    _check_headerpars(fobj, ext)
    hdr = _getheader(fobj, ext)
    names = hdr["WCSNAME*"]
    d = []
    for key in names:
        wkey = key.replace('WCSNAME', '')
        if wkey == '': wkey = ' '
        d.append(wkey)
    return d


def wcsnames(fobj, ext=None):
    """
    Returns a dictionary of wcskey: WCSNAME pairs

    Parameters
    ----------
    fobj : stri, `astropy.io.fits.HDUList` or `astropy.io.fits.Header`
          fits file name, fits file object or fits header
    ext : int or None
         extension number
         if None, fobj must be a header

    """
    _check_headerpars(fobj, ext)
    hdr = _getheader(fobj, ext)
    names = hdr["WCSNAME*"]
    d = {}
    for keyword, value in names.items():
        wkey = keyword.replace('WCSNAME', '')
        if wkey == '': wkey = ' '
        d[wkey] = value
    return d


def available_wcskeys(fobj, ext=None):
    """
    Returns a list of characters which are not used in the header
    with WCSNAME keyword. Any of them can be used to save a new
    WCS.

    Parameters
    ----------
    fobj : str, `astropy.io.fits.HDUList` or `astropy.io.fits.Header`
        fits file name, fits file object or fits header
    ext : int or None
        extension number
        if None, fobj must be a header
    """
    _check_headerpars(fobj, ext)
    hdr = _getheader(fobj, ext)
    all_keys = list(string.ascii_uppercase)
    used_keys = wcskeys(hdr)
    try:
        used_keys.remove(" ")
    except ValueError:
        pass
    [all_keys.remove(key) for key in used_keys]
    return all_keys


def next_wcskey(fobj, ext=None):
    """
    Returns next available character to be used for an alternate WCS

    Parameters
    ----------
    fobj : str, `astropy.io.fits.HDUList` or `astropy.io.fits.Header`
        fits file name, fits file object or fits header
    ext : int or None
        extension number
        if None, fobj must be a header
    """
    _check_headerpars(fobj, ext)
    hdr = _getheader(fobj, ext)
    allkeys = available_wcskeys(hdr)
    if allkeys != []:
        return allkeys[0]
    else:
        return None


def getKeyFromName(header, wcsname):
    """
    If WCSNAME is found in header, return its key, else return
    None. This is used to update an alternate WCS
    repeatedly and not generate new keys every time.

    Parameters
    ----------
    header :  `astropy.io.fits.Header`
    wcsname : str
        value of WCSNAME
    """
    wkey = None
    names = wcsnames(header)
    wkeys = []
    for item in names.items():
        if item[1].lower() == wcsname.lower():
            wkeys.append(item[0])
    wkeys.sort()
    if len(wkeys) > 0:
        wkey = wkeys[-1]
    else:
        wkey = None
    return wkey


def pc2cd(hdr, key=' '):
    """
    Convert a PC matrix to a CD matrix.

    WCSLIB (and PyWCS) recognizes CD keywords as input
    but converts them and works internally with the PC matrix.
    to_header() returns the PC matrix even if the input was a
    CD matrix. To keep input and output consistent we check
    for has_cd and convert the PC back to CD.

    Parameters
    ----------
    hdr: `astropy.io.fits.Header`

    """
    for c in ['1_1', '1_2', '2_1', '2_2']:
        try:
            val = hdr['PC{0}{1}'.format(c, key)]
            del hdr['PC{0}{1}'.format(c, key)]
        except KeyError:
            if c == '1_1' or c == '2_2':
                val = 1.
            else:
                val = 0.
        hdr['CD{0}{1}'.format(c, key)] = val
    return hdr


def _parpasscheck(fobj, ext, wcskey, fromext=None, toext=None, reusekey=False):
    """
    Check input parameters to altwcs functions

    fobj : str or `astropy.io.fits.HDUList` object
        a file name or a file object
    ext : int, a tuple,a python list of integers or a python list
        of tuples (e.g.('sci',1))
        fits extensions to work with
    wcskey : str
        "A"-"Z" or " "- Used for one of 26 alternate WCS definitions
    wcsname : str
        (optional)
        if given and wcskey is " ", will try to restore by WCSNAME value
    reusekey : bool
        A flag which indicates whether to reuse a wcskey in the header
    """
    if not isinstance(fobj, fits.HDUList):
        print("First parameter must be a fits file object or a file name.")
        return False

    # first one covers the case of an object created in memory
    # (e.g. headerlet) for which fileinfo returns None
    if fobj.fileinfo(0) is None:
        pass
    else:
        # an HDUList object with associated file
        if fobj.fileinfo(0)['filemode'] is not 'update':
            print("First parameter must be a file name or a file object opened in 'update' mode.")
            return False

    if not isinstance(ext, int) and not isinstance(ext, tuple) \
        and not isinstance(ext, str) \
        and not isinstance(ext, list) and ext is not None:
        print("Ext must be integer, tuple, string,a list of int extension numbers, \n\
        or a list of tuples representing a fits extension, for example ('sci', 1).")
        return False

    if not isinstance(fromext, str) and fromext is not None:
        print("fromext must be a string representing a valid extname")
        return False

    if not isinstance(toext, list) and not isinstance(toext, str) and \
                        toext is not None:
        print("toext must be a string or a list of strings representing extname")
        return False

    if len(wcskey) != 1:
        print('Parameter wcskey must be a character - one of "A"-"Z" or " "')
        return False

    return True


def closefobj(fname, f):
    """
    Functions in this module accept as input a file name or a file object.
    If the input was a file name (string) we close the object. If the user
    passed a file object we leave it to the user to close it.
    """
    if isinstance(fname, str):
        f.close()


def mapFitsExt2HDUListInd(fname, extname):
    """
    Map FITS extensions with 'EXTNAME' to HDUList indexes.
    """

    if not isinstance(fname, fits.HDUList):
        f = fits.open(fname)
        close_file = True
    else:
        f = fname
        close_file = False
    d = {}
    for hdu in f:
        if 'EXTNAME' in hdu.header and hdu.header['EXTNAME'] == extname:
            extver = hdu.header['EXTVER']
            d[(extname, extver)] = f.index_of((extname, extver))
    if close_file:
        f.close()
    return d



import os

from astropy.io import fits
from stsci.tools import parseinput
from stsci.tools import fileutil
from stsci.tools import teal
from .. import __version__
from .. import updatewcs


allowed_corr_dict = {'vacorr': 'VACorr', 'tddcorr': 'TDDCorr', 'npolcorr': 'NPOLCorr',
                     'd2imcorr': 'DET2IMCorr'}


__taskname__ = __name__.split('.')[-1] # needed for help string
__package__ = updatewcs.__name__
#__version__ = stwcs.__version__

#
#### Interfaces used by TEAL
#
def getHelpAsString(docstring=False):
    """
    return useful help from a file in the script directory called __taskname__.help
    """
    install_dir = os.path.dirname(__file__)
    htmlfile = os.path.join(install_dir, 'htmlhelp', __taskname__ + '.html')
    helpfile = os.path.join(install_dir, __taskname__ + '.help')
    if docstring or (not docstring and not os.path.exists(htmlfile)):
        helpString = __taskname__ + ' Version ' + __version__ + '\n\n'
        if os.path.exists(helpfile):
            helpString += teal.getHelpFileAsString(__taskname__, __file__)
        else:
            helpString += eval('.'.join([__package__, __taskname__, '__doc__']))

    else:
        helpString = 'file://' + htmlfile

    return helpString


def run(configObj=None):

    # Interpret primary parameters from configObj instance
    input = configObj['input']

    # create dictionary of remaining parameters, deleting extraneous ones
    # such as those above
    cdict = configObj.dict()
    # remove any rules defined for the TEAL interface
    if "_RULES_" in cdict: del cdict['_RULES_']
    del cdict['_task_name_']
    del cdict['input']
    del cdict['extname']

    # parse input
    input, altfiles = parseinput.parseinput(configObj['input'])

    # Insure that all input files have a correctly archived
    #    set of OPUS WCS keywords
    # Legacy files from OTFR, like all WFPC2 data from OTFR, will only
    #   have the OPUS WCS keywords archived using a prefix of 'O'
    # These keywords need to be converted to the Paper I alternate WCS
    #   standard using a wcskey (suffix) of 'O'
    # If an alternate WCS with wcskey='O' already exists, this will copy
    #   the values from the old prefix-'O' WCS keywords to insure the correct
    #   OPUS keyword values get archived for use with updatewcs.
    #
    for file in input:
        # Check to insure that there is a valid reference file to be used
        idctab = fits.getval(file, 'idctab')
        if not os.path.exists(fileutil.osfn(idctab)):
            print('No valid distortion reference file ', idctab, ' found in ', file, '!')
            raise ValueError

    # Re-define 'cdict' to only have switches for steps supported by that instrument
    # the set of supported steps are defined by the dictionary
    #    updatewcs.apply_corrections.allowed_corrections
    #
    for file in input:
        # get instrument name from input file
        instr = fits.getval(file, 'INSTRUME')
        # make copy of input parameters dict for this file
        fdict = cdict.copy()
        # Remove any parameter that is not part of this instrument's allowed corrections
        for step in allowed_corr_dict:
            if allowed_corr_dict[step] not in updatewcs.apply_corrections.allowed_corrections[instr]:
                fdict[step]
        # Call 'updatewcs' on correctly archived file
        updatewcs.updatewcs(file, **fdict)



import copy
import numpy as np
from astropy.io import fits

import stwcs
from . import altwcs
from ..updatewcs import utils
from stsci.tools import fileutil

DEFAULT_WCS_KEYS = ['CRVAL1', 'CRVAL2', 'CRPIX1', 'CRPIX2',
                    'CD1_1', 'CD1_2', 'CD2_1', 'CD2_2',
                    'CTYPE1', 'CTYPE2', 'ORIENTAT']
DEFAULT_PRI_KEYS = ['HDRNAME', 'SIPNAME', 'NPOLNAME', 'D2IMNAME', 'DESCRIP']
COL_FITSKW_DICT = {'RMS_RA': 'sci.crder1', 'RMS_DEC': 'sci.crder2',
                   'NMatch': 'sci.nmatch', 'Catalog': 'sci.catalog'}

###
### WCSEXT table related keyword archive functions
###
def init_wcscorr(input, force=False):
    """
    This function will initialize the WCSCORR table if it is not already present,
    and look for WCS keywords with a prefix of 'O' as the original OPUS
    generated WCS as the initial row for the table or use the current WCS
    keywords as initial row if no 'O' prefix keywords are found.

    This function will NOT overwrite any rows already present.

    This function works on all SCI extensions at one time.
    """
    # TODO: Create some sort of decorator or (for Python2.5) context for
    # opening a FITS file and closing it when done, if necessary
    if not isinstance(input, fits.HDUList):
        # input must be a filename, so open as `astropy.io.fits.HDUList` object
        fimg = fits.open(input, mode='update')
        need_to_close = True
    else:
        fimg = input
        need_to_close = False

    # Do not try to generate a WCSCORR table for a simple FITS file
    numsci = fileutil.countExtn(fimg)
    if len(fimg) == 1 or numsci == 0:
        return

    enames = []
    for e in fimg: enames.append(e.name)
    if 'WCSCORR' in enames:
        if not force:
            return
        else:
            del fimg['wcscorr']
    print('Initializing new WCSCORR table for ', fimg.filename())

    used_wcskeys = altwcs.wcskeys(fimg['SCI', 1].header)

    # define the primary columns of the WCSEXT table with initial rows for each
    # SCI extension for the original OPUS solution
    numwcs = len(used_wcskeys)
    if numwcs == 0: numwcs = 1

    # create new table with more rows than needed initially to make it easier to
    # add new rows later
    wcsext = create_wcscorr(descrip=True, numrows=numsci, padding=(numsci * numwcs) + numsci * 4)
    # Assign the correct EXTNAME value to this table extension
    wcsext.header['TROWS'] = (numsci * 2, 'Number of updated rows in table')
    wcsext.header['EXTNAME'] = ('WCSCORR', 'Table with WCS Update history')
    wcsext.header['EXTVER'] = 1

    # define set of WCS keywords which need to be managed and copied to the table
    wcs1 = stwcs.wcsutil.HSTWCS(fimg, ext=('SCI', 1))
    idc2header = True
    if wcs1.idcscale is None:
        idc2header = False
    wcs_keywords = list(wcs1.wcs2header(idc2hdr=idc2header).keys())

    prihdr = fimg[0].header
    prihdr_keys = DEFAULT_PRI_KEYS
    pri_funcs = {'SIPNAME': stwcs.updatewcs.utils.build_sipname,
                 'NPOLNAME': stwcs.updatewcs.utils.build_npolname,
                 'D2IMNAME': stwcs.updatewcs.utils.build_d2imname}

    # Now copy original OPUS values into table
    for extver in range(1, numsci + 1):
        rowind = find_wcscorr_row(wcsext.data,
                                  {'WCS_ID': 'OPUS', 'EXTVER': extver,
                                   'WCS_key': 'O'})
        # There should only EVER be a single row for each extension with OPUS values
        rownum = np.where(rowind)[0][0]
        # print 'Archiving OPUS WCS in row number ',rownum,' in WCSCORR table for SCI,',extver

        hdr = fimg['SCI', extver].header
        # define set of WCS keywords which need to be managed and copied to the table
        if used_wcskeys is None:
            used_wcskeys = altwcs.wcskeys(hdr)
        # Check to see whether or not there is an OPUS alternate WCS present,
        # if so, get its values directly, otherwise, archive the PRIMARY WCS
        # as the OPUS values in the WCSCORR table
        if 'O' not in used_wcskeys:
            altwcs.archiveWCS(fimg, ('SCI', extver), wcskey='O', wcsname='OPUS')
        wkey = 'O'

        wcs = stwcs.wcsutil.HSTWCS(fimg, ext=('SCI', extver), wcskey=wkey)
        wcshdr = wcs.wcs2header(idc2hdr=idc2header)

        if wcsext.data.field('CRVAL1')[rownum] != 0:
            # If we find values for these keywords already in the table, do not
            # overwrite them again
            print('WCS keywords already updated...')
            break
        for key in wcs_keywords:
            if key in wcsext.data.names:
                wcsext.data.field(key)[rownum] = wcshdr[(key + wkey)[:8]]
        # Now get any keywords from PRIMARY header needed for WCS updates
        for key in prihdr_keys:
            if key in prihdr:
                val = prihdr[key]
            else:
                val = ''
            wcsext.data.field(key)[rownum] = val

    # Now that we have archived the OPUS alternate WCS, remove it from the list
    # of used_wcskeys
    if 'O' in used_wcskeys:
        used_wcskeys.remove('O')

    # Now copy remaining alternate WCSs into table
    # TODO: Much of this appears to be redundant with update_wcscorr; consider
    # merging them...
    for uwkey in used_wcskeys:
        for extver in range(1, numsci + 1):
            hdr = fimg['SCI', extver].header
            wcs = stwcs.wcsutil.HSTWCS(fimg, ext=('SCI', extver),
                                       wcskey=uwkey)
            wcshdr = wcs.wcs2header()
            if 'WCSNAME' + uwkey not in wcshdr:
                wcsid = utils.build_default_wcsname(fimg[0].header['idctab'])
            else:
                wcsid = wcshdr['WCSNAME' + uwkey]

            # identify next empty row
            rowind = find_wcscorr_row(wcsext.data,
                                      selections={'wcs_id': ['', '0.0']})
            rows = np.where(rowind)
            if len(rows[0]) > 0:
                rownum = np.where(rowind)[0][0]
            else:
                print('No available rows found for updating. ')

            # Update selection columns for this row with relevant values
            wcsext.data.field('WCS_ID')[rownum] = wcsid
            wcsext.data.field('EXTVER')[rownum] = extver
            wcsext.data.field('WCS_key')[rownum] = uwkey

            # Look for standard WCS keyword values
            for key in wcs_keywords:
                if key in wcsext.data.names:
                    wcsext.data.field(key)[rownum] = wcshdr[key + uwkey]
            # Now get any keywords from PRIMARY header needed for WCS updates
            for key in prihdr_keys:
                if key in pri_funcs:
                    val = pri_funcs[key](fimg)[0]
                else:
                    if key in prihdr:
                        val = prihdr[key]
                    else:
                        val = ''
                wcsext.data.field(key)[rownum] = val

    # Append this table to the image FITS file
    fimg.append(wcsext)
    # force an update now
    # set the verify flag to 'warn' so that it will always succeed, but still
    # tell the user if PyFITS detects any problems with the file as a whole
    utils.updateNEXTENDKw(fimg)

    fimg.flush('warn')

    if need_to_close:
        fimg.close()


def find_wcscorr_row(wcstab, selections):
    """
    Return an array of indices from the table (NOT HDU) 'wcstab' that matches the
    selections specified by the user.

    The row selection criteria must be specified as a dictionary with
    column name as key and value(s) representing the valid desired row values.
    For example, {'wcs_id':'OPUS','extver':2}.
    """

    mask = None
    for i in selections:
        icol = wcstab.field(i)
        if isinstance(icol, np.chararray): icol = icol.rstrip()
        selecti = selections[i]
        if not isinstance(selecti, list):
            if isinstance(selecti, str):
                selecti = selecti.rstrip()
            bmask = (icol == selecti)
            if mask is None:
                mask = bmask.copy()
            else:
                mask = np.logical_and(mask, bmask)
            del bmask
        else:
            for si in selecti:
                if isinstance(si, str):
                    si = si.rstrip()
                bmask = (icol == si)
                if mask is None:
                    mask = bmask.copy()
                else:
                    mask = np.logical_or(mask, bmask)
                del bmask

    return mask


def archive_wcs_file(image, wcs_id=None):
    """
    Update WCSCORR table with rows for each SCI extension to record the
    newly updated WCS keyword values.
    """

    if not isinstance(image, fits.HDUList):
        fimg = fits.open(image, mode='update')
        close_image = True
    else:
        fimg = image
        close_image = False

    update_wcscorr(fimg, wcs_id=wcs_id)

    if close_image:
        fimg.close()


def update_wcscorr(dest, source=None, extname='SCI', wcs_id=None, active=True):
    """
    Update WCSCORR table with a new row or rows for this extension header. It
    copies the current set of WCS keywords as a new row of the table based on
    keyed WCSs as per Paper I Multiple WCS standard).

    Parameters
    ----------
    dest : HDUList
        The HDU list whose WCSCORR table should be appended to (the WCSCORR HDU
        must already exist)
    source : HDUList, optional
        The HDU list containing the extension from which to extract the WCS
        keywords to add to the WCSCORR table.  If None, the dest is also used
        as the source.
    extname : str, optional
        The extension name from which to take new WCS keywords.  If there are
        multiple extensions with that name, rows are added for each extension
        version.
    wcs_id : str, optional
        The name of the WCS to add, as in the WCSNAMEa keyword.  If
        unspecified, all the WCSs in the specified extensions are added.
    active: bool, optional
        When True, indicates that the update should reflect an update of the
        active WCS information, not just appending the WCS to the file as a
        headerlet
    """
    if not isinstance(dest, fits.HDUList):
        dest = fits.open(dest, mode='update')
    fname = dest.filename()

    if source is None:
        source = dest

    if extname == 'PRIMARY':
        return

    numext = fileutil.countExtn(source, extname)
    if numext == 0:
        raise ValueError('No %s extensions found in the source HDU list.'
                         % extname)
    # Initialize the WCSCORR table extension in dest if not already present
    init_wcscorr(dest)
    try:
        dest.index_of('WCSCORR')
    except KeyError:
        return

    # check to see whether or not this is an up-to-date table
    # replace with newly initialized table with current format
    old_table = dest['WCSCORR']
    wcscorr_cols = ['WCS_ID', 'EXTVER', 'SIPNAME',
                    'HDRNAME', 'NPOLNAME', 'D2IMNAME']

    for colname in wcscorr_cols:
        if colname not in old_table.data.columns.names:
            print("WARNING:    Replacing outdated WCSCORR table...")
            outdated_table = old_table.copy()
            del dest['WCSCORR']
            init_wcscorr(dest)
            old_table = dest['WCSCORR']
            break

    # Current implementation assumes the same WCS keywords are in each
    # extension version; if this should not be assumed then this can be
    # modified...
    wcs_keys = altwcs.wcskeys(source[(extname, 1)].header)
    wcs_keys = [kk for kk in wcs_keys if kk]
    if ' ' not in wcs_keys: wcs_keys.append(' ')  # Insure that primary WCS gets used
    # apply logic for only updating WCSCORR table with specified keywords
    # corresponding to the WCS with WCSNAME=wcs_id
    if wcs_id is not None:
        wnames = altwcs.wcsnames(source[(extname, 1)].header)
        wkeys = []
        for letter in wnames:
            if wnames[letter] == wcs_id:
                wkeys.append(letter)
        if len(wkeys) > 1 and ' ' in wkeys:
            wkeys.remove(' ')
        wcs_keys = wkeys
    wcshdr = stwcs.wcsutil.HSTWCS(source, ext=(extname, 1)).wcs2header()
    wcs_keywords = list(wcshdr.keys())

    if 'O' in wcs_keys:
        wcs_keys.remove('O')  # 'O' is reserved for original OPUS WCS

    # create new table for hdr and populate it with the newly updated values
    new_table = create_wcscorr(descrip=True, numrows=0, padding=len(wcs_keys) * numext)
    prihdr = source[0].header

    # Get headerlet related keywords here
    sipname, idctab = utils.build_sipname(source, fname, "None")
    npolname, npolfile = utils.build_npolname(source, None)
    d2imname, d2imfile = utils.build_d2imname(source, None)
    if 'hdrname' in prihdr:
        hdrname = prihdr['hdrname']
    else:
        hdrname = ''

    idx = -1
    for wcs_key in wcs_keys:
        for extver in range(1, numext + 1):
            extn = (extname, extver)
            if 'SIPWCS' in extname and not active:
                tab_extver = 0  # Since it has not been added to the SCI header yet
            else:
                tab_extver = extver
            hdr = source[extn].header
            if 'WCSNAME' + wcs_key in hdr:
                wcsname = hdr['WCSNAME' + wcs_key]
            else:
                wcsname = utils.build_default_wcsname(hdr['idctab'])

            selection = {'WCS_ID': wcsname, 'EXTVER': tab_extver,
                         'SIPNAME': sipname, 'HDRNAME': hdrname,
                         'NPOLNAME': npolname, 'D2IMNAME': d2imname
                         }

            # Ensure that an entry for this WCS is not already in the dest
            # table; if so just skip it
            rowind = find_wcscorr_row(old_table.data, selection)
            if np.any(rowind):
                continue

            idx += 1

            wcs = stwcs.wcsutil.HSTWCS(source, ext=extn, wcskey=wcs_key)
            wcshdr = wcs.wcs2header()

            # Update selection column values
            for key, val in selection.items():
                if key in new_table.data.names:
                    new_table.data.field(key)[idx] = val

            for key in wcs_keywords:
                if key in new_table.data.names:
                    new_table.data.field(key)[idx] = wcshdr[key + wcs_key]

            for key in DEFAULT_PRI_KEYS:
                if key in new_table.data.names and key in prihdr:
                    new_table.data.field(key)[idx] = prihdr[key]
            # Now look for additional, non-WCS-keyword table column data
            for key in COL_FITSKW_DICT:
                fitkw = COL_FITSKW_DICT[key]
                # Interpret any 'pri.hdrname' or
                # 'sci.crpix1' formatted keyword names
                if '.' in fitkw:
                    srchdr, fitkw = fitkw.split('.')
                    if 'pri' in srchdr.lower(): srchdr = prihdr
                    else: srchdr = source[extn].header
                else:
                    srchdr = source[extn].header

                if fitkw + wcs_key in srchdr:
                    new_table.data.field(key)[idx] = srchdr[fitkw + wcs_key]

    # If idx was never incremented, no rows were added, so there's nothing else
    # to do...
    if idx < 0:
        return

    # Now, we need to merge this into the existing table
    rowind = find_wcscorr_row(old_table.data, {'wcs_id': ['', '0.0']})
    old_nrows = np.where(rowind)[0][0]
    new_nrows = new_table.data.shape[0]

    # check to see if there is room for the new row
    if (old_nrows + new_nrows) > old_table.data.shape[0] - 1:
        pad_rows = 2 * new_nrows
        # if not, create a new table with 'pad_rows' new empty rows
        upd_table = fits.BinTableHDU.from_columns(old_table.columns, header=old_table.header,
                                                  nrows=old_table.data.shape[0] + pad_rows)
    else:
        upd_table = old_table
        pad_rows = 0
    # Now, add
    for name in old_table.columns.names:
        if name in new_table.data.names:
            # reset the default values to ones specific to the row definitions
            for i in range(pad_rows):
                upd_table.data.field(name)[old_nrows + i] = old_table.data.field(name)[-1]
            # Now populate with values from new table
            upd_table.data.field(name)[old_nrows:old_nrows + new_nrows] = \
                new_table.data.field(name)
    upd_table.header['TROWS'] = old_nrows + new_nrows

    # replace old extension with newly updated table extension
    dest['WCSCORR'] = upd_table


def restore_file_from_wcscorr(image, id='OPUS', wcskey=''):
    """ Copies the values of the WCS from the WCSCORR based on ID specified by user.
    The default will be to restore the original OPUS-derived values to the Primary WCS.
    If wcskey is specified, the WCS with that key will be updated instead.
    """

    if not isinstance(image, fits.HDUList):
        fimg = fits.open(image, mode='update')
        close_image = True
    else:
        fimg = image
        close_image = False
    numsci = fileutil.countExtn(fimg)
    wcs_table = fimg['WCSCORR']
    orig_rows = (wcs_table.data.field('WCS_ID') == 'OPUS')
    # create an HSTWCS object to figure out what WCS keywords need to be updated
    wcsobj = stwcs.wcsutil.HSTWCS(fimg, ext=('sci', 1))
    wcshdr = wcsobj.wcs2header()
    for extn in range(1, numsci + 1):
        # find corresponding row from table
        ext_rows = (wcs_table.data.field('EXTVER') == extn)
        erow = np.where(np.logical_and(ext_rows, orig_rows))[0][0]
        for key in wcshdr:
            if key in wcs_table.data.names:  # insure that keyword is column in table
                tkey = key

                if 'orient' in key.lower():
                    key = 'ORIENT'
                if wcskey == '':
                    skey = key
                else:
                    skey = key[:7] + wcskey
                fimg['sci', extn].header[skey] = wcs_table.data.field(tkey)[erow]
        for key in DEFAULT_PRI_KEYS:
            if key in wcs_table.data.names:
                if wcskey == '':
                    pkey = key
                else:
                    pkey = key[: 7] + wcskey
                fimg[0].header[pkey] = wcs_table.data.field(key)[erow]

    utils.updateNEXTENDKw(fimg)

    # close the image now that the update has been completed.
    if close_image:
        fimg.close()


def create_wcscorr(descrip=False, numrows=1, padding=0):
    """
    Return the basic definitions for a WCSCORR table.
    The dtype definitions for the string columns are set to the maximum allowed so
    that all new elements will have the same max size which will be automatically
    truncated to this limit upon updating (if needed).

    The table is initialized with rows corresponding to the OPUS solution
    for all the 'SCI' extensions.
    """

    trows = numrows + padding
    # define initialized arrays as placeholders for column data
    # TODO: I'm certain there's an easier way to do this... for example, simply
    # define the column names and formats, then create an empty array using
    # them as a dtype, then create the new table from that array.
    def_float64_zeros = np.array([0.0] * trows, dtype=np.float64)
    def_float64_ones = def_float64_zeros + 1.0
    def_float_col = {'format': 'D', 'array': def_float64_zeros.copy()}
    def_float1_col = {'format': 'D', 'array': def_float64_ones.copy()}
    def_str40_col = {'format': '40A',
                     'array': np.array([''] * trows, dtype='S40')}
    def_str24_col = {'format': '24A',
                     'array': np.array([''] * trows, dtype='S24')}
    def_int32_col = {'format': 'J',
                     'array': np.array([0] * trows, dtype=np.int32)}

    # If more columns are needed, simply add their definitions to this list
    col_names = [('HDRNAME', def_str24_col), ('SIPNAME', def_str24_col),
                 ('NPOLNAME', def_str24_col), ('D2IMNAME', def_str24_col),
                 ('CRVAL1', def_float_col), ('CRVAL2', def_float_col),
                 ('CRPIX1', def_float_col), ('CRPIX2', def_float_col),
                 ('CD1_1', def_float_col), ('CD1_2', def_float_col),
                 ('CD2_1', def_float_col), ('CD2_2', def_float_col),
                 ('CTYPE1', def_str24_col), ('CTYPE2', def_str24_col),
                 ('ORIENTAT', def_float_col), ('PA_V3', def_float_col),
                 ('RMS_RA', def_float_col), ('RMS_Dec', def_float_col),
                 ('NMatch', def_int32_col), ('Catalog', def_str40_col)]

    # Define selector columns
    id_col = fits.Column(name='WCS_ID', format='40A',
                         array=np.array(['OPUS'] * numrows + [''] * padding,
                                        dtype='S24'))
    extver_col = fits.Column(name='EXTVER', format='I',
                             array=np.array(list(range(1, numrows + 1)),
                                            dtype=np.int16))
    wcskey_col = fits.Column(name='WCS_key', format='A',
                             array=np.array(['O'] * numrows + [''] * padding,
                                            dtype='S'))
    # create list of remaining columns to be added to table
    col_list = [id_col, extver_col, wcskey_col]  # start with selector columns

    for c in col_names:
        cdef = copy.deepcopy(c[1])
        col_list.append(fits.Column(name=c[0], format=cdef['format'],
                        array=cdef['array']))

    if descrip:
        col_list.append(
            fits.Column(name='DESCRIP', format='128A',
                        array=np.array(
                            ['Original WCS computed by OPUS'] * numrows,
                            dtype='S128')))

    # Now create the new table from the column definitions
    newtab = fits.BinTableHDU.from_columns(fits.ColDefs(col_list), nrows=trows)
    # The fact that setting .name is necessary should be considered a bug in
    # pyfits.
    # TODO: Make sure this is fixed in pyfits, then remove this
    newtab.name = 'WCSCORR'

    return newtab


def delete_wcscorr_row(wcstab, selections=None, rows=None):
    """
    Sets all values in a specified row or set of rows to default values

    This function will essentially erase the specified row from the table
    without actually removing the row from the table. This avoids the problems
    with trying to resize the number of rows in the table while preserving the
    ability to update the table with new rows again without resizing the table.

    Parameters
    ----------
    wcstab: object
        PyFITS binTable object for WCSCORR table
    selections: dict
        Dictionary of wcscorr column names and values to be used to select
        the row or set of rows to erase
    rows: int, list
        If specified, will specify what rows from the table to erase regardless
        of the value of 'selections'
    """

    if selections is None and rows is None:
        print('ERROR: Some row selection information must be provided!')
        print('       Either a row numbers or "selections" must be provided.')
        raise ValueError

    delete_rows = None
    if rows is None:
        if 'wcs_id' in selections and selections['wcs_id'] == 'OPUS':
            delete_rows = None
            print('WARNING: OPUS WCS information can not be deleted from WCSCORR table.')
            print('         This row will not be deleted!')
        else:
            rowind = find_wcscorr_row(wcstab, selections=selections)
            delete_rows = np.where(rowind)[0].tolist()
    else:
        if not isinstance(rows, list):
            rows = [rows]
        delete_rows = rows

    # Insure that rows pointing to OPUS WCS do not get deleted, even by accident
    for row in delete_rows:
        if wcstab['WCS_key'][row] == 'O' or wcstab['WCS_ID'][row] == 'OPUS':
            del delete_rows[delete_rows.index(row)]

    if delete_rows is None:
        return

    # identify next empty row
    rowind = find_wcscorr_row(wcstab, selections={'wcs_id': ['', '0.0']})
    last_blank_row = np.where(rowind)[0][-1]

    # copy values from blank row into user-specified rows
    for colname in wcstab.names:
        wcstab[colname][delete_rows] = wcstab[colname][last_blank_row]


def update_wcscorr_column(wcstab, column, values, selections=None, rows=None):
    """
    Update the values in 'column' with 'values' for selected rows

    Parameters
    ----------
    wcstab: object
        PyFITS binTable object for WCSCORR table
    column: string
        Name of table column with values that need to be updated
    values: string, int, or list
        Value or set of values to copy into the selected rows for the column
    selections: dict
        Dictionary of wcscorr column names and values to be used to select
        the row or set of rows to erase
    rows: int, list
        If specified, will specify what rows from the table to erase regardless
        of the value of 'selections'
    """
    if selections is None and rows is None:
        print('ERROR: Some row selection information must be provided!')
        print('       Either a row numbers or "selections" must be provided.')
        raise ValueError

    if not isinstance(values, list):
        values = [values]

    update_rows = None
    if rows is None:
        if 'wcs_id' in selections and selections['wcs_id'] == 'OPUS':
            update_rows = None
            print('WARNING: OPUS WCS information can not be deleted from WCSCORR table.')
            print('         This row will not be deleted!')
        else:
            rowind = find_wcscorr_row(wcstab, selections=selections)
            update_rows = np.where(rowind)[0].tolist()
    else:
        if not isinstance(rows, list):
            rows = [rows]
        update_rows = rows

    if update_rows is None:
        return

    # Expand single input value to apply to all selected rows
    if len(values) > 1 and len(values) < len(update_rows):
        print('ERROR: Number of new values', len(values))
        print('       does not match number of rows', len(update_rows), ' to be updated!')
        print('       Please enter either 1 value or the same number of values')
        print('       as there are rows to be updated.')
        print('    Table will not be updated...')
        raise ValueError

    if len(values) == 1 and len(values) < len(update_rows):
        values = values * len(update_rows)
    # copy values from blank row into user-specified rows
    for row in update_rows:
        wcstab[column][row] = values[row]



import os
from stsci.tools import teal, parseinput

from .. import __version__
from ..wcsutil import headerlet

__taskname__ = __name__.split('.')[-1]  # needed for help string
__package__ = headerlet.__name__
# __version__ = stwcs.__version__


############### Interfaces used by TEAL ###############

def getHelpAsString(docstring=False):
    """
    return useful help from a file in the script directory called __taskname__.help
    """
    install_dir = os.path.dirname(__file__)
    htmlfile = os.path.join(install_dir, 'htmlhelp', __taskname__ + '.html')
    helpfile = os.path.join(install_dir, __taskname__ + '.help')
    if docstring or (not docstring and not os.path.exists(htmlfile)):
        helpString = __taskname__ + ' Version ' + __version__ + '\n\n'
        if os.path.exists(helpfile):
            helpString += teal.getHelpFileAsString(__taskname__, __file__)
    else:
        helpString = 'file://' + htmlfile

    return helpString


def run(configObj=None):

    # start by interpreting filename and hdrlet inputs
    filename = parseinput.parseinput(configObj['filename'])[0]
    hdrlet = parseinput.parseinput(configObj['hdrlet'])[0]

    if configObj['primary']:
        # Call function with properly interpreted input parameters
        # Syntax: apply_headerlet_as_primary(filename, hdrlet, attach=True,
        #            archive=True, force=False, verbose=False)
        headerlet.apply_headerlet_as_primary(filename,
                                             hdrlet, attach=configObj['attach'],
                                             archive=configObj['archive'], force=configObj['force'],
                                             logging=configObj['logging'])
    else:
        wcsname = configObj['wcsname']
        if wcsname in ['', ' ', 'INDEF']: wcsname = None
        wcskey = configObj['wcskey']
        if wcskey == '': wcskey = None
        # Call function with properly interpreted input parameters
        #         apply_headerlet_as_alternate(filename, hdrlet, attach=True,
        #                        wcskey=None, wcsname=None, verbose=False)
        headerlet.apply_headerlet_as_alternate(filename,
                                               hdrlet, attach=configObj['attach'],
                                               wcsname=wcsname, wcskey=wcskey,
                                               logging=configObj['logging'])



import logging
import time

import numpy as np
from astropy.io import fits

from stsci.tools import fileutil

logger = logging.getLogger('stwcs.updatewcs.npol')


class NPOLCorr(object):
    """
    Defines a Lookup table prior distortion correction as per WCS paper IV.
    It uses a reference file defined by the NPOLFILE (suffix 'NPL') keyword
    in the primary header.

    Notes
    -----
    - Using extensions in the reference file create a WCSDVARR extensions
      and add them to the science file.
    - Add record-valued keywords to the science extension header to describe
      the lookup tables.
    - Add a keyword 'NPOLEXT' to the science extension header to store
      the name of the reference file used to create the WCSDVARR extensions.

    If WCSDVARR extensions exist and `NPOLFILE` is different from `NPOLEXT`,
    a subsequent update will overwrite the existing extensions.
    If WCSDVARR extensions were not found in the science file, they will be added.

    It is assumed that the NPL reference files were created to work with IDC tables
    but will be applied with SIP coefficients. A transformation is applied to correct
    for the fact that the lookup tables will be applied before the first order coefficients
    which are in the CD matrix when the SIP convention is used.
    """

    def updateWCS(cls, fobj):
        """
        Parameters
        ----------
        fobj : `astropy.io.fits.HDUList` object
            Science file, for which a distortion correction in a NPOLFILE is available

        """
        logger.info("\n\tStarting NPOL: %s" % time.asctime())
        try:
            assert isinstance(fobj, fits.HDUList)
        except AssertionError:
            logger.exception('\n\tInput must be a fits.HDUList object')
            raise

        cls.applyNPOLCorr(fobj)
        nplfile = fobj[0].header['NPOLFILE']

        new_kw = {'NPOLEXT': nplfile}
        return new_kw

    updateWCS = classmethod(updateWCS)

    def applyNPOLCorr(cls, fobj):
        """
        For each science extension in a fits file object:
            - create a WCSDVARR extension
            - update science header
            - add/update NPOLEXT keyword
        """
        nplfile = fileutil.osfn(fobj[0].header['NPOLFILE'])
        # Map WCSDVARR EXTVER numbers to extension numbers
        wcsdvarr_ind = cls.getWCSIndex(fobj)
        for ext in fobj:
            try:
                extname = ext.header['EXTNAME'].lower()
            except KeyError:
                continue
            if extname == 'sci':
                extversion = ext.header['EXTVER']
                ccdchip = cls.get_ccdchip(fobj, extname='SCI', extver=extversion)
                header = ext.header
                # get the data arrays from the reference file and transform
                # them for use with SIP
                dx, dy = cls.getData(nplfile, ccdchip)
                idccoeffs = cls.getIDCCoeffs(header)

                if idccoeffs is not None:
                    dx, dy = cls.transformData(dx, dy, idccoeffs)

                # Determine EXTVER for the WCSDVARR extension from the
                # NPL file (EXTNAME, EXTVER) kw.
                # This is used to populate DPj.EXTVER kw
                wcsdvarr_x_version = 2 * extversion - 1
                wcsdvarr_y_version = 2 * extversion
                for ename in zip(['DX', 'DY'], [wcsdvarr_x_version, wcsdvarr_y_version], [dx, dy]):
                    error_val = ename[2].max()
                    cls.addSciExtKw(header, wdvarr_ver=ename[1], npol_extname=ename[0], error_val=error_val)
                    hdu = cls.createNpolHDU(header, npolfile=nplfile,
                                            wdvarr_ver=ename[1], npl_extname=ename[0],
                                            data=ename[2], ccdchip=ccdchip)
                    if wcsdvarr_ind:
                        fobj[wcsdvarr_ind[ename[1]]] = hdu
                    else:
                        fobj.append(hdu)

    applyNPOLCorr = classmethod(applyNPOLCorr)

    def getWCSIndex(cls, fobj):

        """
        If fobj has WCSDVARR extensions:
            returns a mapping of their EXTVER kw to file object extension numbers
        if fobj does not have WCSDVARR extensions:
            an empty dictionary is returned
        """
        wcsd = {}
        for e in range(len(fobj)):
            try:
                ename = fobj[e].header['EXTNAME']
            except KeyError:
                continue
            if ename == 'WCSDVARR':
                wcsd[fobj[e].header['EXTVER']] = e
        logger.debug("A map of WSCDVARR externsions %s" % wcsd)
        return wcsd

    getWCSIndex = classmethod(getWCSIndex)

    def addSciExtKw(cls, hdr, wdvarr_ver=None, npol_extname=None, error_val=0.0):
        """
        Adds kw to sci extension to define WCSDVARR lookup table extensions

        """
        if npol_extname == 'DX':
            j = 1
        else:
            j = 2

        cperror = 'CPERR%s' % j
        cpdis = 'CPDIS%s' % j
        dpext = 'DP%s.' % j + 'EXTVER'
        dpnaxes = 'DP%s.' % j + 'NAXES'
        dpaxis1 = 'DP%s.' % j + 'AXIS.1'
        dpaxis2 = 'DP%s.' % j + 'AXIS.2'
        keys = [cperror, cpdis, dpext, dpnaxes, dpaxis1, dpaxis2]
        values = {cperror: error_val,
                  cpdis: 'Lookup',
                  dpext: wdvarr_ver,
                  dpnaxes: 2,
                  dpaxis1: 1,
                  dpaxis2: 2}

        comments = {cperror: 'Maximum error of NPOL correction for axis %s' % j,
                    cpdis: 'Prior distortion function type',
                    dpext: 'Version number of WCSDVARR extension containing lookup distortion table',
                    dpnaxes: 'Number of independent variables in distortion function',
                    dpaxis1: 'Axis number of the jth independent variable in a distortion function',
                    dpaxis2: 'Axis number of the jth independent variable in a distortion function'
                    }
        # Look for HISTORY keywords. If present, insert new keywords before them
        before_key = 'HISTORY'
        if before_key not in hdr:
            before_key = None

        for key in keys:
            hdr.set(key, value=values[key], comment=comments[key], before=before_key)

    addSciExtKw = classmethod(addSciExtKw)

    def getData(cls, nplfile, ccdchip):
        """
        Get the data arrays from the reference NPOL files
        Make sure 'CCDCHIP' in the npolfile matches "CCDCHIP' in the science file.
        """
        npl = fits.open(nplfile)
        for ext in npl:
            nplextname  = ext.header.get('EXTNAME', "")
            nplccdchip  = ext.header.get('CCDCHIP', 1)
            if nplextname == 'DX' and nplccdchip == ccdchip:
                xdata = ext.data.copy()
                continue
            elif nplextname == 'DY' and nplccdchip == ccdchip:
                ydata = ext.data.copy()
                continue
            else:
                continue
        npl.close()
        return xdata, ydata
    getData = classmethod(getData)

    def transformData(cls, dx, dy, coeffs):
        """
        Transform the NPOL data arrays for use with SIP
        """
        ndx, ndy = np.dot(coeffs, [dx.ravel(), dy.ravel()]).astype(np.float32)
        ndx.shape = dx.shape
        ndy.shape = dy.shape
        return ndx, ndy

    transformData = classmethod(transformData)

    def getIDCCoeffs(cls, header):
        """
        Return a matrix of the scaled first order IDC coefficients.
        """
        try:
            ocx10 = header['OCX10']
            ocx11 = header['OCX11']
            ocy10 = header['OCY10']
            ocy11 = header['OCY11']
            coeffs = np.array([[ocx11, ocx10], [ocy11, ocy10]], dtype=np.float32)
        except KeyError:
            logger.exception('\n\tFirst order IDCTAB coefficients are not available. \n\
            Cannot convert SIP to IDC coefficients.')
            return None
        try:
            idcscale = header['IDCSCALE']
        except KeyError:
            logger.exception("IDCSCALE not found in header - setting it to 1.")
            idcscale = 1

        return np.linalg.inv(coeffs / idcscale)

    getIDCCoeffs = classmethod(getIDCCoeffs)

    def createNpolHDU(cls, sciheader, npolfile=None, wdvarr_ver=1, npl_extname=None,
                      data=None, ccdchip=1):
        """
        Creates an HDU to be added to the file object.
        """
        hdr = cls.createNpolHdr(sciheader, npolfile=npolfile, wdvarr_ver=wdvarr_ver,
                                npl_extname=npl_extname, ccdchip=ccdchip)
        hdu = fits.ImageHDU(header=hdr, data=data)
        return hdu

    createNpolHDU = classmethod(createNpolHDU)

    def createNpolHdr(cls, sciheader, npolfile, wdvarr_ver, npl_extname, ccdchip):
        """
        Creates a header for the WCSDVARR extension based on the NPOL reference file
        and sci extension header. The goal is to always work in image coordinates
        (also for subarrays and binned images. The WCS for the WCSDVARR extension
        i ssuch that a full size npol table is created and then shifted or scaled
        if the science image is a subarray or binned image.
        """
        npl = fits.open(npolfile)
        npol_phdr = npl[0].header
        for ext in npl:
            try:
                nplextname = ext.header['EXTNAME']
                nplextver = ext.header['EXTVER']
            except KeyError:
                continue
            nplccdchip = cls.get_ccdchip(npl, extname=nplextname, extver=nplextver)
            if nplextname == npl_extname and nplccdchip == ccdchip:
                npol_header = ext.header
                break
            else:
                continue
        npl.close()

        naxis = npl[1].header['NAXIS']
        ccdchip = nplextname  # npol_header['CCDCHIP']

        kw = {'NAXIS': 'Size of the axis',
              'CDELT': 'Coordinate increment along axis',
              'CRPIX': 'Coordinate system reference pixel',
              'CRVAL': 'Coordinate system value at reference pixel',
              }

        kw_comm1 = {}
        kw_val1 = {}
        for key in kw.keys():
            for i in range(1, naxis + 1):
                si = str(i)
                kw_comm1[key + si] = kw[key]

        for i in range(1, naxis + 1):
            si = str(i)
            kw_val1['NAXIS' + si] = npol_header.get('NAXIS' + si)
            kw_val1['CDELT' + si] = npol_header.get('CDELT' + si, 1.0) * \
                sciheader.get('LTM' + si + '_' + si, 1)
            kw_val1['CRPIX' + si] = npol_header.get('CRPIX' + si, 0.0)
            kw_val1['CRVAL' + si] = (npol_header.get('CRVAL' + si, 0.0) -
                                     sciheader.get('LTV' + str(i), 0))

        kw_comm0 = {'XTENSION': 'Image extension',
                    'BITPIX': 'IEEE floating point',
                    'NAXIS': 'Number of axes',
                    'EXTNAME': 'WCS distortion array',
                    'EXTVER': 'Distortion array version number',
                    'PCOUNT': 'Special data area of size 0',
                    'GCOUNT': 'One data group',
                    }

        kw_val0 = {'XTENSION': 'IMAGE',
                   'BITPIX': -32,
                   'NAXIS': naxis,
                   'EXTNAME': 'WCSDVARR',
                   'EXTVER': wdvarr_ver,
                   'PCOUNT': 0,
                   'GCOUNT': 1,
                   'CCDCHIP': ccdchip,
                   }
        cdl = []
        for key in kw_comm0.keys():
            cdl.append((key, kw_val0[key], kw_comm0[key]))
        for key in kw_comm1.keys():
            cdl.append((key, kw_val1[key], kw_comm1[key]))
        # Now add keywords from NPOLFILE header to document source of calibration
        # include all keywords after and including 'FILENAME' from header
        start_indx = -1
        end_indx = 0
        for i, c in enumerate(npol_phdr):
            if c == 'FILENAME':
                start_indx = i
            if c == '':  # remove blanks from end of header
                end_indx = i + 1
                break
        if start_indx >= 0:
            for card in npol_phdr.cards[start_indx: end_indx]:
                cdl.append(card)

        hdr = fits.Header(cards=cdl)

        return hdr

    createNpolHdr = classmethod(createNpolHdr)

    def get_ccdchip(cls, fobj, extname, extver):
        """
        Given a science file or npol file determine CCDCHIP
        """
        ccdchip = 1
        if fobj[0].header['INSTRUME'] == 'ACS' and fobj[0].header['DETECTOR'] == 'WFC':
            ccdchip = fobj[extname, extver].header['CCDCHIP']
        elif fobj[0].header['INSTRUME'] == 'WFC3' and fobj[0].header['DETECTOR'] == 'UVIS':
            ccdchip = fobj[extname, extver].header['CCDCHIP']
        elif fobj[0].header['INSTRUME'] == 'WFPC2':
            ccdchip = fobj[extname, extver].header['DETECTOR']
        elif fobj[0].header['INSTRUME'] == 'STIS':
            ccdchip = fobj[extname, extver].header['DETECTOR']
        elif fobj[0].header['INSTRUME'] == 'NICMOS':
            ccdchip = fobj[extname, extver].header['CAMERA']
        return ccdchip

    get_ccdchip = classmethod(get_ccdchip)
